#  _________________________________________________________________________
#
#  Coopr: A COmmon Optimization Python Repository
#  Copyright (c) 2008 Sandia Corporation.
#  This software is distributed under the BSD License.
#  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
#  the U.S. Government retains certain rights in this software.
#  For more information, see the Coopr README.txt file.
#  _________________________________________________________________________

__all__ = ['Suffix','active_export_suffix_generator','active_import_suffix_generator']

import sys
import logging
import weakref
import copy
from six import iteritems, iterkeys, itervalues
import pyutilib.math

from coopr.pyomo.base.component import Component, register_component

logger = logging.getLogger('coopr.pyomo')

# A list of convenient suffix generators, including:
#   - active_export_suffix_generator **(used by problem writers)
#   - export_suffix_generator
#   - active_import_suffix_generator **(used by OptSolver and PyomoModel._load_solution)
#   - import_suffix_generator
#   - active_local_suffix_generator
#   - local_suffix_generator
#   - active_suffix_generator
#   - suffix_generator

def active_export_suffix_generator(a_block,datatype=False):
    if (datatype is False):
        for name, suffix in iteritems(a_block.active_subcomponents(Suffix)):
            if suffix.exportEnabled() is True:
                yield name, suffix
    else:
        for name, suffix in iteritems(a_block.active_subcomponents(Suffix)):
            if (suffix.exportEnabled() is True) and \
               (suffix.getDatatype() is datatype):
                yield name, suffix

def export_suffix_generator(a_block,datatype=False):
    if (datatype is False):
        for name, suffix in iteritems(a_block.subcomponents(Suffix)):
            if suffix.exportEnabled() is True:
                yield name, suffix
    else:
        for name, suffix in iteritems(a_block.subcomponents(Suffix)):
            if (suffix.exportEnabled() is True) and \
               (suffix.getDatatype() is datatype):
                yield name, suffix

def active_import_suffix_generator(a_block,datatype=False):
    if (datatype is False):
        for name, suffix in iteritems(a_block.active_subcomponents(Suffix)):
            if suffix.importEnabled() is True:
                yield name, suffix
    else:
        for name, suffix in iteritems(a_block.active_subcomponents(Suffix)):
            if (suffix.importEnabled() is True) and \
               (suffix.getDatatype() is datatype):
                yield name, suffix

def import_suffix_generator(a_block,datatype=False):
    if (datatype is False):
        for name, suffix in iteritems(a_block.subcomponents(Suffix)):
            if suffix.importEnabled() is True:
                yield name, suffix
    else:
        for name, suffix in iteritems(a_block.subcomponents(Suffix)):
            if (suffix.importEnabled() is True) and \
               (suffix.getDatatype() is datatype):
                yield name, suffix

def active_local_suffix_generator(a_block,datatype=False):
    if (datatype is False):
        for name, suffix in iteritems(a_block.active_subcomponents(Suffix)):
            if suffix.getDirection() is Suffix.LOCAL:
                yield name, suffix
    else:
        for name, suffix in iteritems(a_block.active_subcomponents(Suffix)):
            if (suffix.getDirection() is Suffix.LOCAL) and \
               (suffix.getDatatype() is datatype):
                yield name, suffix

def local_suffix_generator(a_block,datatype=False):
    if (datatype is False):
        for name, suffix in iteritems(a_block.subcomponents(Suffix)):
            if suffix.getDirection() is Suffix.LOCAL:
                yield name, suffix
    else:
        for name, suffix in iteritems(a_block.subcomponents(Suffix)):
            if (suffix.getDirection() is Suffix.LOCAL) and \
               (suffix.getDatatype() is datatype):
                yield name, suffix

def active_suffix_generator(a_block,datatype=False):
    if (datatype is False):
        for name, suffix in iteritems(a_block.active_subcomponents(Suffix)):
            yield name, suffix
    else:
        for name, suffix in iteritems(a_block.active_subcomponents(Suffix)):
            if suffix.getDatatype() is datatype:
                yield name, suffix

def suffix_generator(a_block,datatype=False):
    if (datatype is False):
        for name, suffix in iteritems(a_block.subcomponents(Suffix)):
            yield name, suffix
    else:
        for name, suffix in iteritems(a_block.subcomponents(Suffix)):
            if suffix.getDatatype() is datatype:
                yield name, suffix

# THOUGHTS: 

# TODO: Should we, when setting the value on a component, start collecting
#       a dictionary of components by their type (like on Blocks)

# TODO: Should calling clearValue on a parent component (like _VarArray) clear
#       its own dictionary entry AS WELL AS any possible dictionary entries
#       corresponding to its subcomponent entries (e.g. _VarData members)?
#       Right now this does not happen. One could ask the same of setValue.


class Suffix(Component):
    """A model suffix, representing extranious model data"""

    """
    Constructor Arguments:
        direction   The direction of information flow for this suffix.
                        By default, this is LOCAL, indicating that no
                        suffix data is exported or imported.
        datatype    A variable type associated with all values of this
                        suffix.
    """

    # Suffix Directions
    # If more directions are added be sure to update the error
    # message in the setDirection method
    LOCAL  = 0 # neither
    EXPORT = 1 # sent to solver or other external location
    IMPORT = 2 # obtained from solver or other external source
    IMPORT_EXPORT = 3 # both
    SuffixDirections = (LOCAL,EXPORT,IMPORT,IMPORT_EXPORT)
    SuffixDirectionToStr = {LOCAL:'Suffix.LOCAL',EXPORT:'Suffix.EXPORT',IMPORT:'Suffix.IMPORT',IMPORT_EXPORT:'Suffix.IMPORT_EXPORT'}
    # Suffix Datatypes
    FLOAT = 4
    INT = 0
    SuffixDatatypes = (FLOAT,INT,None)
    SuffixDatatypeToStr = {FLOAT:'Suffix.FLOAT',INT:'Suffix.INT',None:str(None)}

    def __getstate__(self):
        """
        This method must be defined for deepcopy/pickling because 
        this class relies on component ids.
        """
        result = super(Suffix, self).__getstate__()
        result['_subcomponent_value_map'] = tuple((ref(),val) for ref,val in itervalues(self._subcomponent_value_map) if ref() is not None)
        return result

    def __setstate__(self, state):
        """
        This method must be defined for deepcopy/pickling because 
        this class relies on component ids.
        """
        id_func = id
        weakref_func = weakref.ref
        state['_subcomponent_value_map'] = dict((id_func(comp),(weakref_func(comp),val)) for comp,val in state['_subcomponent_value_map'])
        return super(Suffix,self).__setstate__(state)

    def __init__(self, *args, **kwds):
        
        # Suffix type information
        self._direction = None
        self._datatype = None
        self._rule = None
        
        # The meat of suffixes... a dictionary mapping 
        # Pyomo component ids to suffix values
        self._subcomponent_value_map = {}
        
        # TODO: __FIX_ME__: Blocks assume everything is a (Sparse)IndexedComponent inside
        #       __setattr__ so until this is fixed we need to define a dummy _index data
        #       member
        self._index = None
        
        # The suffix direction
        direction = kwds.pop('direction',Suffix.LOCAL)
        # The suffix datatype
        datatype = kwds.pop('datatype',Suffix.FLOAT)
        # The suffix construction rule
        self._rule = kwds.pop('rule',None)
        
        # Initialize base class
        kwds.setdefault('ctype', Suffix)
        Component.__init__(self, *args, **kwds)
        
        # Check that keyword values make sense 
        # (these function have internal error checking).
        self.setDirection(direction)
        self.setDatatype(datatype) 

        if self._rule is None:
            self.construct()

    def reset(self):
        """
        Reconstructs this component by clearing all values and re-calling
        construction rule if it exists.
        """
        self.clearAllValues()
        self._constructed = False
        self.construct()

    def construct(self, data=None):
        """
        Constructs this component, applying rule if it exists.
        """
        generate_debug_messages = (__debug__ is True) and (logger.isEnabledFor(logging.DEBUG) is True)

        if generate_debug_messages is True:
            logger.debug("Constructing suffix %s",self.cname())

        if self._constructed is True:
            return

        self._constructed = True

        if self._rule is not None:
            # TODO: expande=?
            self.updateValues(self._rule(self._parent()))

    def exportEnabled(self):
        """
        Returns True when this suffix is enabled for export to solvers.
        """
        return bool(self._direction & Suffix.EXPORT)

    def importEnabled(self):
        """
        Returns True when this suffix is enabled for import from solutions.
        """
        return bool(self._direction & Suffix.IMPORT)

    def updateValues(self,data_buffer,expand=True):
        """
        Updates the suffix data given a list of component,value tuples. Provides
        an improvement in efficiency over calling setValue on every component.
        """
        if expand is True:
            # set array components first
            id_func = id
            weakref_func = weakref.ref
            # TODO: Check if we need to copy in order to iterate over twice (e.g., it's a generator or iterator)
            data_buffer = list(data_buffer)
            for comp,value in ((_comp,_value) for _comp,_value in data_buffer if _comp.component() is _comp):
                self._subcomponent_value_map.update((id_func(subcomp),(weakref_func(subcomp),value)) for subcomp in itervalues(comp))
            # set subcomponents last
            self._subcomponent_value_map.update((id_func(comp),(weakref_func(comp),val)) for comp,val in data_buffer if comp.component() is not comp)
        else:
            id_func = id
            weakref_func = weakref.ref
            self._subcomponent_value_map.update((id_func(comp),(weakref_func(comp),val)) for comp,val in data_buffer)

    def extractValues(self):
        """
        Extract all data stored on this Suffix into a list of component, value tuples.
        """
        return list((subcomponent_ref(),value) for subcomponent_ref,value in itervalues(self._subcomponent_value_map) if subcomponent_ref() is not None)
        
    def setValue(self,component,value,expand=True):
        """
        Sets the value of this suffix on the specified component.
        """
        if expand is True:
            if component.component() is not component:
                self._subcomponent_value_map[id(component)] = (weakref.ref(component),value)
            else:
                id_func = id
                weakref_func = weakref.ref
                self._subcomponent_value_map.update((id_func(subcomponent),(weakref_func(subcomponent),value)) for subcomponent in itervalues(component))
        else:
            self._subcomponent_value_map[id(component)] = (weakref.ref(component),value)

    def setAllValues(self,value):
        """
        Sets the value of this suffix on all components.
        """
        self._subcomponent_value_map = dict((ID,(ref,value)) for ID,(ref,old_value) in iteritems(self._subcomponent_value_map))

    def getValue(self,component):
        """
        Returns the current value of this suffix for the specified component.
        """
        # check if the component has been assigned an individual
        # suffix value.
        # **Note: We do this safely by checking that the id is still valid
        #         for the original component it was assigned to by checking
        #         the weakref
        cid = id(component)
        if cid in self._subcomponent_value_map:
            ref,value = self._subcomponent_value_map[cid]
            if ref() is component:
                return value
            else:
                # A rare case where the id was reassigned
                del self._subcomponent_value_map[cid]
        return None

    def clearValue(self,component, expand=True):
        """
        Clears suffix information for a component.
        """
        # clearing the value on a specified component
        if (expand is False) or (component.component() is not component):
            cid = id(component)
            if cid in self._subcomponent_value_map:
                del self._subcomponent_value_map[cid]
        else:
            for subcomponent in itervalues(component):
                cid = id(subcomponent)
                if cid in self._subcomponent_value_map:
                    del self._subcomponent_value_map[cid]

    def clearAllValues(self):
        """
        Clears all suffix data.
        """
        self._subcomponent_value_map = {}
        
    def setDatatype(self,datatype):
        """
        Set the suffix datatype.
        """
        if datatype not in self.SuffixDatatypes:
            raise ValueError("Suffix datatype must be one of: %s. \n" \
                              "Value given: %s" % (list(Suffix.SuffixDatatypeToStr.values()),datatype))
        self._datatype = datatype

    def getDatatype(self):
        """
        Return the suffix datatype.
        """
        return self._datatype

    def setDirection(self,direction):
        """ 
        Set the suffix direction.
        """
        if not direction in self.SuffixDirections:
            raise ValueError("Suffix direction must be one of: %s. \n" \
                              "Value given: %s" % (list(self.SuffixDirectionToStr.values()),direction))
        self._direction = direction

    def getDirection(self):
        """
        Return the suffix direction.
        """
        return self._direction

    def __str__(self):
        """
        Return a string representation of the suffix.  If the name attribute is None,
        then return ''
        """
        if self.cname() is None:
            return ""
        else:
            return self.cname()

    def _pprint(self):
        return (
            [('Direction', self.SuffixDirectionToStr[self._direction]),
             ('Datatype', self.SuffixDatatypeToStr[self._datatype]),
             ],
            ( (str(k()),v) for k,v in itervalues(self._subcomponent_value_map) 
              if k() is not None ),
            ("Key","Value"),
            lambda k,v: [ k, v ]
            )

    def Xpprint(self, ostream=None, verbose=False, prefix='  '):
        """
        Print component information. Setting verbose to True includes suffix values."""
        if ostream is None:
            ostream = sys.stdout
        ostream.write(prefix+self.cname()+" :")
        if not self.doc is None:
            ostream.write(self.doc+'\n')
        else:
            ostream.write('\n')
        ostream.write(prefix)
        ostream.write("\tDirection= ")
        if self._direction in self.SuffixDirectionToStr:
            ostream.write(self.SuffixDirectionToStr[self._direction]+"\n")
        else:
            raise ValueError("Unexpected Suffix direction encountered in pprint method: %s" % (self._direction))
        ostream.write("\tDatatype= ")
        if self._datatype in self.SuffixDatatypeToStr:
            ostream.write(self.SuffixDatatypeToStr[self._datatype]+"\n")
        else:
            raise ValueError("Unexpected Suffix datatype encountered in pprint method: %s" % (self._datatype))
        # print all component values
        if verbose is True:
            ostream.write('\tValues=\n')
            name_data = dict((ref().cname(True),val) for ref,val \
                                                     in itervalues(self._subcomponent_value_map) \
                                                     if ref() is not None)
            for key in sorted(iterkeys(name_data)):
                ostream.write('\t  %s = %s\n' % (key, name_data[key]))

register_component(Suffix, "Extraneous model data")



#
# Test that EXPORT type Suffix components work with the NL writer
#

import os
import sys
import copy
import itertools
from os.path import abspath, dirname, join
currdir = dirname(abspath(__file__))

import coopr.environ
from coopr.pyomo import *
from coopr.opt import ProblemFormat
import pyutilib.th as unittest

class TestSuffix(unittest.TestCase):

    # test that EXPORT suffixes on variables,
    # constraints, objectives, and models
    # will end up in the NL file with integer tags
    def test_EXPORT_suffixes_int(self):
        model = ConcreteModel()
        model.junk = Suffix(direction=Suffix.EXPORT,datatype=Suffix.INT)
        model.junk_inactive = Suffix(direction=Suffix.EXPORT,datatype=Suffix.INT)

        model.x = Var()
        model.junk.setValue(model.x,1)
        model.junk_inactive.setValue(model.x,1)

        model.y = Var([1,2])
        model.junk.setValue(model.y,2)
        model.junk_inactive.setValue(model.y,2)

        model.obj = Objective(expr=model.x+summation(model.y))
        model.junk.setValue(model.obj,3)
        model.junk_inactive.setValue(model.obj,3)

        model.conx = Constraint(expr=model.x>=1)
        model.junk.setValue(model.conx,4)
        model.junk_inactive.setValue(model.conx,4)

        model.cony = Constraint([1,2],rule=lambda model,i: model.y[i]>=1)
        model.junk.setValue(model.cony,5)
        model.junk_inactive.setValue(model.cony,5)

        model.junk.setValue(model,6)
        model.junk_inactive.setValue(model,6)

        # This one should NOT end up in the NL file
        model.junk_inactive.deactivate()

        model.write(filename=join(currdir,"EXPORT_suffixes.test.nl"),
                    format=ProblemFormat.nl,
                    io_options={"symbolic_solver_labels" : False})

        self.assertFileEqualsBaseline(join(currdir,"EXPORT_suffixes.test.nl"),
                                      join(currdir,"EXPORT_suffixes_int.baseline.nl"))

    # test that EXPORT suffixes on variables,
    # constraints, objectives, and models
    # will end up in the NL file with floating point tags
    def test_EXPORT_suffixes_float(self):
        model = ConcreteModel()
        model.junk = Suffix(direction=Suffix.EXPORT,datatype=Suffix.FLOAT)
        model.junk_inactive = Suffix(direction=Suffix.EXPORT,datatype=Suffix.FLOAT)

        model.x = Var()
        model.junk.setValue(model.x,1)
        model.junk_inactive.setValue(model.x,1)

        model.y = Var([1,2])
        model.junk.setValue(model.y,2)
        model.junk_inactive.setValue(model.y,2)

        model.obj = Objective(expr=model.x+summation(model.y))
        model.junk.setValue(model.obj,3)
        model.junk_inactive.setValue(model.obj,3)

        model.conx = Constraint(expr=model.x>=1)
        model.junk.setValue(model.conx,4)
        model.junk_inactive.setValue(model.conx,4)

        model.cony = Constraint([1,2],rule=lambda model,i: model.y[i]>=1)
        model.junk.setValue(model.cony,5)
        model.junk_inactive.setValue(model.cony,5)

        model.junk.setValue(model,6)
        model.junk_inactive.setValue(model,6)

        # This one should NOT end up in the NL file
        model.junk_inactive.deactivate()

        model.write(filename=join(currdir,"EXPORT_suffixes.test.nl"),
                    format=ProblemFormat.nl,
                    io_options={"symbolic_solver_labels" : False})

        self.assertFileEqualsBaseline(join(currdir,"EXPORT_suffixes.test.nl"),
                                      join(currdir,"EXPORT_suffixes_float.baseline.nl"))

    # Test that user defined ref suffixes fail to
    # merge with those created from translating the SOSConstraint
    # component when variables get assigned duplicate values for ref
    def test_EXPORT_suffixes_with_SOSConstraint_duplicateref(self):
        model = ConcreteModel()
        model.ref = Suffix(direction=Suffix.EXPORT,datatype=Suffix.INT)
        model.y = Var([1,2,3])
        model.obj = Objective(expr=summation(model.y))

        # The NL writer will convert this constraint to ref and sosno
        # suffixes on model.y
        model.sos_con = SOSConstraint(var=model.y, set=[1,2,3], sos=1)

        for i,val in zip([1,2,3],[11,12,13]):
            model.ref.setValue(model.y[i],val)
        
        try:
            model.write(filename=join(currdir,"junk.nl"),
                        format=ProblemFormat.nl,
                        io_options={"symbolic_solver_labels" : False})
        except RuntimeError:
            pass
        else:
            os.remove(join(currdir,"junk.nl"))
            self.fail("The NL writer should have thrown an exception "\
                      "when overlap of SOSConstraint generated suffixes "\
                      "and user declared suffixes occurs.")
        
        try:
            os.remove(join(currdir,"junk.nl"))
        except:
            pass

    # Test that user defined sosno suffixes fail to
    # merge with those created from translating the SOSConstraint
    # component when variables get assigned duplicate values for sosno
    def test_EXPORT_suffixes_with_SOSConstraint_duplicatesosno(self):
        model = ConcreteModel()
        model.sosno = Suffix(direction=Suffix.EXPORT,datatype=Suffix.INT)
        model.y = Var([1,2,3])
        model.obj = Objective(expr=summation(model.y))

        # The NL writer will convert this constraint to ref and sosno
        # suffixes on model.y
        model.sos_con = SOSConstraint(var=model.y, set=[1,2,3], sos=1)

        for i in [1,2,3]:
            model.sosno.setValue(model.y[i],-1)
        
        try:
            model.write(filename=join(currdir,"junk.nl"),
                        format=ProblemFormat.nl,
                        io_options={"symbolic_solver_labels" : False})
        except RuntimeError:
            pass
        else:
            os.remove(join(currdir,"junk.nl"))
            self.fail("The NL writer should have thrown an exception "\
                      "when overlap of SOSConstraint generated suffixes "\
                      "and user declared suffixes occurs.")
        try:
            os.remove(join(currdir,"junk.nl"))
        except:
            pass

    # Test that user defined sosno suffixes fail to
    # merge with those created from translating the SOSConstraint
    # component when variables get assigned duplicate values for sosno
    def test_EXPORT_suffixes_no_datatype(self):
        model = ConcreteModel()
        model.sosno = Suffix(direction=Suffix.EXPORT,datatype=None)
        model.y = Var([1,2,3])
        model.obj = Objective(expr=summation(model.y))

        # The NL writer will convert this constraint to ref and sosno
        # suffixes on model.y
        model.sos_con = SOSConstraint(var=model.y, set=[1,2,3], sos=1)

        for i in [1,2,3]:
            model.sosno.setValue(model.y[i],-1)
        
        try:
            model.write(filename=join(currdir,"junk.nl"),
                        format=ProblemFormat.nl,
                        io_options={"symbolic_solver_labels" : False})
        except RuntimeError:
            pass
        else:
            os.remove(join(currdir,"junk.nl"))
            self.fail("The NL writer should have thrown an exception "\
                      "when using an EXPORT suffix with datatype=None")
        try:
            os.remove(join(currdir,"junk.nl"))
        except:
            pass

if __name__ == "__main__":
    unittest.main()



import types

import pyutilib.th as unittest
import coopr.environ
from coopr.pyomo import *

from coopr.pyomo.tests.ampl.helper import MockFixedValue

_campl_available = False
#FIXME: Disabling C AMPL tests until we decide whether to keep C AMPL module
#       around and keep it up to date with nl writer changes. (ZBF)
#try:
#    import coopr.pyomo.ampl.cAmpl as cAmpl
#    _campl_available = True
#    cgar = coopr.pyomo.ampl.cAmpl.generate_ampl_repn
#    gar = coopr.pyomo.ampl.ampl.py_generate_ampl_repn
#except ImportError:
from coopr.pyomo.expr.ampl_repn import _generate_ampl_repn as gar
from coopr.pyomo.expr.ampl_repn import AmplRepn

class _GenericAmplRepnEqualityTests(unittest.TestCase):
    def setUp(self):
        # Helper function for param init
        def q_initialize(model, i):
            return [2,3,5,7,11,13,17,19,23,29][i-1]
        # Need a model so that variables can be distinguishable
        self.model = ConcreteModel()
        self.model.s = RangeSet(10)
        self.model.p = Param(default=42)
        self.model.q = Param(self.model.s, initialize=q_initialize)
        self.model.w = Var(self.model.s)
        self.model.x = Var()
        self.model.y = Var()
        self.model.z = Var(self.model.s)
        
    def tearDown(self):
        self.model = None

    def assertAmplRepnMatch(self, rep1, rep2):
        self.assertEqual(rep1, rep2)
        self.assertEqual(rep1.is_constant(), rep2.is_constant())
        self.assertEqual(rep1.is_linear(), rep2.is_linear())
        self.assertEqual(rep1.is_nonlinear(), rep2.is_nonlinear())
        self.assertEqual(rep1.needs_sum(), rep2.needs_sum())

class AmplRepnEqualityTests(_GenericAmplRepnEqualityTests):
    """Serves as a test class for the AmplRepn.__eq__ method."""

    def testBasicEquality(self):
        self.assertEqual(AmplRepn(), AmplRepn())

    def testBasicInequality(self):
        self.assertNotEqual(AmplRepn(), None)

    def testVarEquality(self):
        self.assertEqual(gar(self.model.x), gar(self.model.x))

    def testVarInequality(self):
        self.assertNotEqual(gar(self.model.x), gar(self.model.y))

    def testVarCoefInequality(self):
        self.assertNotEqual(gar(self.model.x), gar(2.0 * self.model.x))

    def testFixedValueEquality(self):
        self.assertEqual(gar(MockFixedValue()), gar(MockFixedValue()))

    def testFixedValueInequality(self):
        self.assertNotEqual(gar(MockFixedValue(1)), gar(MockFixedValue(2)))

    def testProductEquality(self):
        expr = self.model.x * self.model.y
        self.assertEqual(gar(expr), gar(expr))

    def testProductInequality(self):
        e1 = self.model.x * self.model.x
        e2 = self.model.y * self.model.y
        self.assertNotEqual(gar(e1), gar(e2))

    def testSumEquality(self):
        expr = self.model.x + self.model.y
        self.assertEqual(gar(expr), gar(expr))

    def testSumInequality(self):
        e1 = self.model.x + self.model.y
        e2 = self.model.x + self.model.x
        self.assertNotEqual(gar(e1), gar(e2))

    def testSumCommutativeEquality(self):
        e1 = self.model.x + self.model.y
        e2 = self.model.y + self.model.x
        self.assertEqual(gar(e1), gar(e2))

    def testPowEquality(self):
        expr = self.model.x ** 2
        self.assertEqual(gar(expr), gar(expr))

    def testPowInequality(self):
        e1 = self.model.x ** 2
        e2 = self.model.x ** 3
        self.assertNotEqual(gar(e1), gar(e2))

    def testIntrinsicEquality(self):
        fns = [sin, cos, tan, sinh, cosh, tanh, asin, acos, atan, asinh, acosh, atanh, log, exp]
        for fn in fns:
            expr = fn(self.model.x)
            self.assertEqual(gar(expr), gar(expr))

    def testIntrinsicInequality(self):
        e1 = sin(self.model.x)
        e2 = cos(self.model.x)
        self.assertNotEqual(gar(e1), gar(e2))

    def testCompoundEquality(self):
        expr = self.model.x + self.model.y + self.model.x * self.model.y
        self.assertEqual(gar(expr), gar(expr))

    def testMoreCompoundEquality(self):
        expr = ((self.model.x + self.model.y) * self.model.x) ** 2
        self.assertEqual(gar(expr), gar(expr))

    def testCompoundCoefficientInequality(self):
        e1 = self.model.x * self.model.y
        e2 = 2.0 * self.model.x * self.model.y
        self.assertNotEqual(gar(e1), gar(e2))
    
    def testQuotientEquality(self):
        expr = self.model.x / self.model.y
        self.assertEqual(gar(expr), gar(expr))

    def testCompoundQuotientEquality(self):
        expr = self.model.y + self.model.x / self.model.y + self.model.y ** 2
        self.assertEqual(gar(expr), gar(expr))

    def testQuotientInequality(self):
        e1 = self.model.x / self.model.y
        e2 = self.model.y / self.model.x
        self.assertNotEqual(gar(e1), gar(e2))

    def testSumEquality(self):
        expr = sum(self.model.z[i] for i in self.model.s)
        self.assertEqual(gar(expr), gar(expr))

    def testCompoundSumEquality(self):
        expr = sum(self.model.z[i] for i in self.model.s) + self.model.x / self.model.y
        self.assertEqual(gar(expr), gar(expr))

@unittest.skipUnless(_campl_available, "C AMPL module required")
class CAmplEqualityCompatTests(_GenericAmplRepnEqualityTests):
    """Tests whether the Python and cAmpl implementations produce identical repns."""

    def testEnvironment(self):
        # Simply ensure the import environment is working.
        with self.assertRaises(ValueError) as cm:
            cAmpl.generate_ampl_repn(None, None, None)

    def testEnvironmentTypes(self):
        self.assertEqual(type(gar), types.FunctionType)
        self.assertEqual(type(cgar), types.BuiltinFunctionType)

    def testFixedValue(self):
        expr = MockFixedValue()
        self.assertAmplRepnMatch(gar(expr), cgar(expr))

    def testVar(self):
        expr = self.model.x
        self.assertAmplRepnMatch(gar(expr), cgar(expr))

    def testProduct(self):
        expr = self.model.x * self.model.y
        self.assertAmplRepnMatch(gar(expr), cgar(expr))

    def testSum(self):
        expr = self.model.x + self.model.y
        self.assertAmplRepnMatch(gar(expr), cgar(expr))

    def testSumCommutative(self):
        e1 = self.model.x + self.model.y
        e2 = self.model.y + self.model.x
        self.assertAmplRepnMatch(gar(e1), cgar(e2))

    def testPow(self):
        expr = self.model.x ** 2
        self.assertAmplRepnMatch(gar(expr), cgar(expr))

    def testIntrinsic(self):
        fns = [sin, cos, tan, sinh, cosh, tanh, asin, acos, atan, asinh, acosh, atanh, log, exp]
        for fn in fns:
            expr = fn(self.model.x)
            self.assertAmplRepnMatch(gar(expr), cgar(expr))

    def testCompound(self):
        expr = self.model.x + self.model.y + self.model.x * self.model.y
        self.assertAmplRepnMatch(gar(expr), cgar(expr))

    def testMoreCompound(self):
        expr = ((self.model.x + self.model.y) * self.model.x) ** 2
        self.assertAmplRepnMatch(gar(expr), cgar(expr))

    def testQuotient(self):
        expr = self.model.x / self.model.y
        self.assertAmplRepnMatch(gar(expr), cgar(expr))

    def testCompoundQuotient(self):
        expr = self.model.y + self.model.x / self.model.y + self.model.y ** 2
        self.assertAmplRepnMatch(gar(expr), cgar(expr))

    def testSum(self):
        expr = sum(self.model.z[i] for i in self.model.s)
        self.assertAmplRepnMatch(gar(expr), cgar(expr))

    def testCompoundSum(self):
        expr = sum(self.model.z[i] for i in self.model.s) + self.model.x / self.model.y
        self.assertAmplRepnMatch(gar(expr), cgar(expr))

    def testSumExpression(self):
        expr = sum(self.model.w[i] / self.model.z[i] ** 3 for i in self.model.s)
        self.assertAmplRepnMatch(gar(expr), cgar(expr))

    def testSumExpressionParam(self):
        expr = sum(value(self.model.q[i]) / self.model.z[i] ** 3 for i in self.model.s)
        self.assertAmplRepnMatch(gar(expr), cgar(expr))

    def testCantilvrConstraintExpr(self):
        # originally from coopr.data.cute cantilvr model.
        expr = sum(value(self.model.q[i]) / self.model.z[i] ** 3 for i in self.model.s) - 1.0
        self.assertAmplRepnMatch(gar(expr), cgar(expr))

    def testCantilvrObjective(self):
        # originally from coopr.data.cute cantilvr model.
        # exposes problem in linear product handling, if present.
        expr = sum(self.model.z[i] for i in self.model.s) * 0.0624
        self.assertAmplRepnMatch(gar(expr), cgar(expr))

if __name__ == "__main__":
    unittest.main(verbosity=2)



# Author: Hans Pirnay, 2012-11-27
# The parametric.mod example from sIPOPT
#
# This Pyomo example is formulated as a python script.
# To run this script execute the following command:
#
# $ coopr_python sipopt_parametric.py
#
# Execution of this script requires that the ipopt_sens
# solver (distributed with Ipopt) is in the current search
# path for executables on this system.

import coopr.environ
from coopr.pyomo import *
from coopr.opt import SolverFactory

### Create the ipopt_sens solver plugin using the ASL interface
solver = 'ipopt_sens'
solver_io = 'nl'
stream_solver = False    # True prints solver output to screen
keepfiles =     False    # True prints intermediate file names (.nl,.sol,...)
opt = SolverFactory(solver,solver_io=solver_io)
###

if opt is None:
    print("")
    print("ERROR: Unable to create solver plugin for 'ipopt_sens'")
    print("")
    exit(1)

### Set this data
nominal_eta1   = 4.5
perturbed_eta1 = 4.0
nominal_eta2   = 1.0
perturbed_eta2 = 1.0

### Create the model
model = ConcreteModel()
# variables
model.x1 = Var(initialize=0.15, within=NonNegativeReals)
model.x2 = Var(initialize=0.15, within=NonNegativeReals)
model.x3 = Var(initialize=0.0, within=NonNegativeReals)
# parameters
model.eta1 = Var()
model.eta2 = Var()
# constraints + objective
model.const1 = Constraint(expr=6*model.x1+3*model.x2+2*model.x3 - model.eta1 == 0)
model.const2 = Constraint(expr=model.eta2*model.x1+model.x2-model.x3-1 == 0)
model.cost   = Objective(expr=model.x1**2 + model.x2**2 + model.x3**2)
model.consteta1 = Constraint(expr=model.eta1 == nominal_eta1)
model.consteta2 = Constraint(expr=model.eta2 == nominal_eta2)
###

### declare suffixes
model.sens_state_0 = Suffix(direction=Suffix.EXPORT)
model.sens_state_1 = Suffix(direction=Suffix.EXPORT)
model.sens_state_value_1 = Suffix(direction=Suffix.EXPORT)
model.sens_sol_state_1  = Suffix(direction=Suffix.IMPORT)
model.sens_init_constr  = Suffix(direction=Suffix.EXPORT)
###

### set sIPOPT data
opt.options['run_sens'] = 'yes'
model.sens_state_0.setValue(model.eta1, 1)
model.sens_state_1.setValue(model.eta1, 1)
model.sens_state_value_1.setValue(model.eta1, perturbed_eta1)
model.sens_state_0.setValue(model.eta2, 2)
model.sens_state_1.setValue(model.eta2, 2)
model.sens_state_value_1.setValue(model.eta2, perturbed_eta2)
model.sens_init_constr.setValue(model.consteta1, 1)
model.sens_init_constr.setValue(model.consteta2, 2)
###

### Send the model to ipopt_sens and collect the solution
results = opt.solve(model, keepfiles=keepfiles, tee=stream_solver)
# load the results (including any values for previously declared
# IMPORT / IMPORT_EXPORT Suffix components)
model.load(results)
###

### Print Solution
print("Nominal and perturbed solution:")
for v in [model.x1, model.x2, model.x3, model.eta1, model.eta2]:
    print("%5s %14g %14g" % (v, value(v), model.sens_sol_state_1.getValue(v)))
###



#
# Symmetric TSP with points on a plane
#
import re
import math
from coopr.pyomo import *

def pyomo_create_model(options=None, model_options=None):
    model = ConcreteModel()
    #
    # Read from the 'points' file
    #
    filename = 'points' if options.points is None else options.points
    INPUT = open(filename, 'r')
    N = int(INPUT.readline())
    x = []
    y = []
    for line in INPUT:
        line = line.strip()
        tokens = re.split('[ /t]+', line)
        x.append(int(tokens[1]))
        y.append(int(tokens[2]))
    INPUT.close()
    #
    # Data initialized from file
    #
    #
    # Number of points
    model.N = Param(within=PositiveIntegers, initialize=N)
    #
    # Index set for points
    model.POINTS = RangeSet(1,model.N)
    #
    # (x,y) location
    def x_rule(model, i):
        return x[i-1]
    model.x = Param(model.POINTS)
    def y_rule(model, i):
        return y[i-1]
    model.y = Param(model.POINTS)
    #
    # Derived data
    #
    #
    # All points are connected
    def LINKS_rule(model):
        return set([(i,j) for i in model.POINTS for j in model.POINTS if i<j])
    model.LINKS = Set(dimen=2)
    #
    # Distance between points
    def d_rule(model, i, j):
        return math.sqrt( (model.x[i]-model.x[j])**2 + (model.y[i]-model.y[j])**2 )
    model.d = Param(model.LINKS)
    #
    # Variables
    #
    model.Z = Var(model.LINKS, within=Binary)
    #
    # Constraints
    #
    #
    # Minimize tour length
    model.tour_length = Objective(expr=summation(model.d, model.Z))
    #
    # Each vertex has degree 2
    def Degrees_rule(model, i):
        return sum(model.Z[i,j] for (i_,j) in model.LINKS if i == i_) + \
            sum(model.Z[j,i] for (j,i_) in model.LINKS if i == i_) == 2
    model.Degrees = Constraint(model.POINTS)

    #
    # NOTE: subtour constraints are added dynamically
    #
    ## Number of subtour elimination cuts
    #model.M = Param(within=PositiveIntegers)
    ## Number of subtour elimination cuts appended
    #model.numcut = Param(within=NonNegativeIntegers)
    #model.CUTINDICES = RangeSet(1,model.M)
    #model.CUTSET = Set(model.CUTINDICES)
    return model



@pyomo_callback('cut-callback')
def cut_callback(solver, model):
    print "Adding cuts"





#  _________________________________________________________________________
#
#  Coopr: A COmmon Optimization Python Repository
#  Copyright (c) 2008 Sandia Corporation.
#  This software is distributed under the BSD License.
#  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
#  the U.S. Government retains certain rights in this software.
#  For more information, see the Coopr README.txt file.
#  _________________________________________________________________________

__all__ = ['Model', 'ConcreteModel', 'AbstractModel', 'global_option']

import array
import copy
import logging
import re
import sys
import traceback
import weakref
import gc
import time
import math
import functools

try:
    from collections import OrderedDict
except ImportError:
    from ordereddict import OrderedDict
try:
    from pympler import muppy
    from pympler import summary
    pympler_available = True
except ImportError:
    pympler_available = False

from six import itervalues, iteritems, StringIO
from six.moves import xrange
try:
    unicode
except:
    basestring = unicode = str

from coopr.core.plugin import ExtensionPoint
from pyutilib.math import *
from pyutilib.misc import quote_split, tuplize, Container, PauseGC, Bunch

import coopr.core
import coopr.opt
from coopr.opt import ProblemFormat, ResultsFormat, guess_format
from coopr.opt.results import SolutionMap, SolverResults, Solution
from coopr.opt.results.container import MapContainer,UndefinedData

from coopr.pyomo.base.var import _VarData, Var
from coopr.pyomo.base.constraint import _ConstraintData
from coopr.pyomo.base.objective import _ObjectiveData
from coopr.pyomo.base.set_types import *
from coopr.pyomo.base.suffix import active_import_suffix_generator
from coopr.pyomo.base.symbol_map import SymbolMap
from coopr.pyomo.base.sparse_indexed_component import SparseIndexedComponent

from coopr.pyomo.base.connector import ConnectorExpander

from coopr.pyomo.base.PyomoModelData import ModelData
from coopr.pyomo.base.DataPortal import *
from coopr.pyomo.base.plugin import *
from coopr.pyomo.base.numvalue import *
from coopr.pyomo.base.block import SimpleBlock
from coopr.pyomo.base.sets import Set
from coopr.pyomo.base.component import register_component, Component

from coopr.pyomo.base.plugin import IModelTransformation, TransformationFactory
logger = logging.getLogger('coopr.pyomo')


def global_option(function, name, value):
    """
    Declare the default value for a global Pyomo configuration option.

    Example use:

    @global_option('config.foo.bar', 1)
    def functor():
        ...
    """
    PyomoConfig._option[tuple(name.split('.'))] = value
    def wrapper_function(*args, **kwargs):
        return function(*args, **kwargs)
    return wrapper_function


class PyomoConfig(Container):
    """
    This is a pyomo-specific configuration object, which is a subclass of Container.
    """

    _option = {}

    def __init__(self, *args, **kw):
        Container.__init__(self, *args, **kw)
        self.set_name('PyomoConfig')
        #
        # Create the nested options specified by the the PyomoConfig._option
        # dictionary, which has been populated with the global_option decorator.
        #
        for item in PyomoConfig._option:
            d = self
            for attr in item[:-1]:
                if not attr in d:
                    d[attr] = Container()
                d = d[attr]
            d[item[-1]] = PyomoConfig._option[item]


class ModelTransformationWrapper(object):

    def __init__(self, model):
        self._model = weakref.ref(model)

    def set_model(self, model):
        self._model = weakref.ref(model)

    def __call__(self, name, **kwds):
        return self.apply(name, **kwds)

    def apply(self, name, **kwds):
        xfrm = TransformationFactory(name)
        if xfrm is None:
            raise ValueError("Bad model transformation '%s'" % name)
        return xfrm(self._model(), **kwds)

    def __dir__(self):
        return TransformationFactory.services()

    def __getattr__(self, name):
        if name.startswith('_'):
            return self.__dict__[name]
        xfrm = TransformationFactory(name)
        if xfrm is None:
            raise ValueError("Bad model transformation '%s'" % name)
        return functools.partial(xfrm, self._model())

    def __setattr__(self, name, val):
        if name == '_model':
            self.__dict__[name] = val
        else:
            raise KeyError("Can only set _model attribute: "+name)

    def __getstate__(self):
        return { '_model': self._model() }

    def __setstate__(self, state):
        self._model = weakref.ref(state['_model'])

    def __copy__(self):
        return type(self)(self._model())

    def __deepcopy__(self, memo):
        return type(self)(self._model())


class Model(SimpleBlock):
    """
    An optimization model.  By default, this defers construction of components
    until data is loaded.
    """

    preprocessor_ep = ExtensionPoint(IPyomoPresolver)


    def __init__ ( self, name='unknown', _deprecate=True, **kwargs ):
        """Constructor"""
        if _deprecate:
            msg = "Using the 'Model' class is deprecated.  Please use the "    \
                  "'AbstractModel' class instead."
            logger.warning( msg )
        #
        # NOTE: The 'ctype' keyword argument is not defined here.  Thus,
        # a model is treated as a 'Block' class type.  This simplifies
        # the definition of the all_blocks() method, since we treat
        # Model and Block objects as the same.  Similarly, this avoids
        # the requirement to import PyomoModel.py in the block.py file.
        #
        SimpleBlock.__init__(self, **kwargs)
        self.name=name
        self.statistics = Container()
        self.config = PyomoConfig()
        self.config.preprocessor = 'pyomo.model.simple_preprocessor'
        self.transform = ModelTransformationWrapper(self)

    def model(self):
        # Special case: the "Model" is always the top-level block, so if
        # this is the top-level block, it must be the model
        if self.parent() is None:
            return self
        else:
            return super(Model, self).model()

    def nvariables(self):
        #if self.statistics.number_of_variables is None:
        #    self.statistics.number_of_variables = len(self.variables())
        return self.statistics.number_of_variables

    #def variables(self):
        #return self.subcomponents(Var)

    def nconstraints(self):
        #if self.statistics.number_of_constraints is None:
        #    self.statistics.number_of_constraints = len(self.constraints())
        return self.statistics.number_of_constraints

    #def constraints(self):
        #return self.subcomponents(Constraint)

    def nobjectives(self):
        #if self.statistics.number_of_objectives is None:
        #    self.statistics.number_of_objectives = len(self.objectives())
        return self.statistics.number_of_objectives

    #def objectives(self):
        #return self.subcomponents(Objective)

    def valid_problem_types(self):
        """This method allows the coopr.opt convert function to work with a Model object."""
        return [ProblemFormat.pyomo]

    def create(self, filename=None, **kwargs):
        """
        Create a concrete instance of this Model, possibly using data
        read in from a file.
        """
        kwargs['filename'] = filename
        functor = kwargs.pop('functor', None)
        if functor is None:
            data = coopr.core.CooprAPIFactory(self.config.create_functor)(self.config, model=self, **kwargs)
        else:
            data = coopr.core.CooprAPIFactory(functor)(self.config, model=self, **kwargs)

        # Creating a model converts it from Abstract -> Concrete
        data.instance._constructed = True
        # Update transformation handle
        data.instance.transform.set_model(data.instance)
        return data.instance

    def clone(self):
        """Create a copy of this model"""
        # FYI: we used to remove all _parent() weakrefs before
        # deepcopying and then restore them on the original and cloned
        # model.  It turns out that this was completely unnecessary and
        # wasteful.  

        # It is unsafe (and currently unsupported) to clone a sub-block
        # in a model
        assert self._parent is None

        # Monkey-patch for deepcopying weakrefs
        # Only required on Python <= 2.6
        #
        if sys.version_info[0] == 2 and sys.version_info[1] <= 6:
            copy._copy_dispatch[weakref.ref] = copy._copy_immutable
            copy._deepcopy_dispatch[weakref.ref] = copy._deepcopy_atomic
            copy._deepcopy_dispatch[weakref.KeyedRef] = copy._deepcopy_atomic

            def dcwvd(self, memo):
                """Deepcopy implementation for WeakValueDictionary class"""
                from copy import deepcopy
                new = self.__class__()
                for key, wr in self.data.items():
                    o = wr()
                    if o is not None:
                        new[deepcopy(key, memo)] = o
                return new
            weakref.WeakValueDictionary.__copy__ = weakref.WeakValueDictionary.copy
            weakref.WeakValueDictionary.__deepcopy__ = dcwvd

            def dcwkd(self, memo):
                """Deepcopy implementation for WeakKeyDictionary class"""
                from copy import deepcopy
                new = self.__class__()
                for key, value in self.data.items():
                    o = key()
                    if o is not none:
                        new[o] = deepcopy(value, memo)
                return new
            weakref.WeakKeyDictionary.__copy__ = weakref.WeakKeyDictionary.copy
            weakref.WeakKeyDictionary.__deepcopy__ = dcwkd
        #
        # Actually do the copy
        #
        return copy.deepcopy(self)

    def reset(self):
        # TODO: check that this works recursively for nested models
        for obj in itervalues(self.subcomponents()):
            obj.reset()

    def preprocess(self, preprocessor=None):
        """Apply the preprocess plugins defined by the user"""
        suspend_gc = PauseGC()
        if preprocessor is None:
            preprocessor = self.config.preprocessor
        coopr.core.CooprAPIFactory(preprocessor)(self.config, model=self)

    #
    # this method is a hack, used strictly by the pyomo command-line utility to
    # allow for user-readable names to be present in solver results objects.
    # this should be removed in the very near future, when labels are isolated
    # completely to solver plugins. in that situation, only human-readable
    # names will be present in a SolverResults object, and this method can
    # be removed.
    #
    # WEH - I think we need this method, even if the labeling method changes.
    #
    # The var/con/obj data is ordered by variable name, sorted by index.
    # This may differ from the declaration order, as well as the
    # instance order.  I haven't figured out how to efficiently generate
    # the declaration order.
    #
    # JDS - I think I agree with JP that this is no longer necessary
    #
    def update_results(self, results):
        results_symbol_map = results._symbol_map
        same_instance = results_symbol_map is not None and \
            results_symbol_map.instance() is self
        new_results = SolverResults()
        new_results.problem = results.problem
        new_results.solver = results.solver

        tmp_name_dict = {}

        for i in xrange(len(results.solution)):

            input_soln = results.solution(i+1)
            input_soln_variable = input_soln.variable
            input_soln_constraint = input_soln.constraint

            new_soln = Solution()
            new_soln.gap = input_soln.gap
            new_soln.status = input_soln.status
            #
            # Variables
            #
            vars = OrderedDict()
            var_id_index_map = dict()
            tmp = {}
            for label, entry in iteritems(input_soln_variable):
                # NOTE: the following is a hack, to handle the ONE_VAR_CONSTANT variable
                if label == "ONE_VAR_CONSTANT":
                    continue
                # translate the label first if there is a symbol map
                # associated with the input solution.
                if same_instance:
                    var_value = results_symbol_map.getObject(label)
                elif results_symbol_map is not None:
                    var_value = results_symbol_map.getEquivalentObject(label, self)
                else:
                    raise RuntimeError("Cannot update from results missing a symbol map")

                if var_value is SymbolMap.UnknownSymbol:
                    msg = "Variable with label '%s' is not in model '%s'."
                    raise KeyError(msg % ( label, self.name, ))

                var_value_id = id(var_value)
                if var_value_id not in var_id_index_map:
                    component = var_value.component()
                    var_id_index_map.update(component.id_index_map())
                var_value_index = var_id_index_map[var_value_id]

                if var_value_index.__class__ is tuple:
                    tmp[(var_value.component().cname(),)+var_value_index] = (var_value.cname(False, tmp_name_dict), entry)
                else:
                    tmp[(var_value.component().cname(),var_value_index)] = (var_value.cname(False, tmp_name_dict), entry)
            for key in sorted(tmp.keys()):
                value = tmp[key]
                vars[value[0]] = value[1]
            new_soln.variable = vars
            #
            # Constraints
            #
            tmp = {}
            con_id_index_map = dict()
            for label, entry in iteritems(input_soln_constraint):
                # NOTE: the following is a hack, to handle the ONE_VAR_CONSTANT variable
                if label == "c_e_ONE_VAR_CONSTANT":
                    continue
                if same_instance:
                    con_value = results_symbol_map.getObject(label)
                elif results_symbol_map is not None:
                    con_value = results_symbol_map.getEquivalentObject(label, self)
                else:
                    raise RuntimeError("Cannot update from results missing a symbol map")

                if con_value is SymbolMap.UnknownSymbol:
                    msg = "Constraint with label '%s' is not in model '%s'."
                    raise KeyError(msg % ( label, self.name, ))

                con_value_id = id(con_value)
                if con_value_id not in con_id_index_map:
                    component = con_value.component()
                    con_id_index_map.update(component.id_index_map())
                con_value_index = con_id_index_map[con_value_id]

                if con_value_index.__class__ is tuple:
                    tmp[(con_value.component().cname(),)+con_value_index] = (con_value.cname(False, tmp_name_dict), entry)
                else:
                    tmp[(con_value.component().cname(),con_value_index)] = (con_value.cname(False, tmp_name_dict), entry)
            for key in sorted(tmp.keys()):
                value = tmp[key]
                new_soln.constraint[value[0]] = value[1]
            #
            # Objectives
            #
            tmp = {}
            for label in input_soln.objective.keys():
                if same_instance:
                    obj_value = results_symbol_map.getObject(label)
                elif results_symbol_map is not None:
                    obj_value = results_symbol_map.getEquivalentObject(label, self)
                else:
                    raise RuntimeError("Cannot update from results missing a symbol map")

                if obj_value is SymbolMap.UnknownSymbol:
                    msg = "Objective with label '%s' is not in model '%s'."
                    raise KeyError(msg % ( label, self.name, ))

                entry = input_soln.objective[label]
                if obj_value.index().__class__ is tuple:
                    tmp[(obj_value.component().cname(),)+obj_value.index()] = (obj_value.cname(False, tmp_name_dict), entry)
                else:
                    tmp[(obj_value.component().cname(),obj_value.index())] = (obj_value.cname(False, tmp_name_dict), entry)
            for key in sorted(tmp.keys()):
                value = tmp[key]
                new_soln.objective.declare(value[0])
                dict.__setitem__(new_soln.objective, value[0], value[1])
            #
            new_results.solution.insert(new_soln)
        return new_results

    def store(self, dp, components, namespace=None):
        for c in components:
            try:
                name = c.cname()
            except:
                name = c
            dp._data.get(namespace,{})[name] = c.data()

    def load(self, arg, namespaces=[None], symbol_map=None,
             allow_consistent_values_for_fixed_vars=False,
             comparison_tolerance_for_fixed_vars=1e-5,
             profile_memory=0, report_timing=False, 
             ignore_invalid_labels=False, id=0):
        """
        Load the model with data from a file or a Solution object
        """

        if arg is None or type(arg) is str:
            self._load_model_data(DataPortal(filename=arg,model=self), namespaces, profile_memory=profile_memory, report_timing=report_timing)
            return True
        elif type(arg) is ModelData:
            self._load_model_data(arg, namespaces, profile_memory=profile_memory, report_timing=report_timing)
            return True
        elif type(arg) is DataPortal:
            self._load_model_data(arg, namespaces, profile_memory=profile_memory, report_timing=report_timing)
            return True
        elif type(arg) is dict:
            self._load_model_data(DataPortal(data_dict=arg,model=self), namespaces, profile_memory=profile_memory, report_timing=report_timing)
            return True
        elif type(arg) is coopr.opt.SolverResults:
            # set the "stale" flag of each variable in the model prior to loading the
            # solution, so you known which variables have "real" values and which ones don't.
            self.flag_vars_as_stale()
            
            # if the solver status not one of either OK or Warning, then error.
            if (arg.solver.status != coopr.opt.SolverStatus.ok) and \
               (arg.solver.status != coopr.opt.SolverStatus.warning):

                if (arg.solver.status == coopr.opt.SolverStatus.aborted) and (len(arg.solution) > 0):
                   print("WARNING - Loading a SolverResults object with an 'aborted' status, but containing a solution")
                else:
                   msg = 'Cannot load a SolverResults object with bad status: %s'
                   raise ValueError(msg % str( arg.solver.status ))

            # but if there is a warning, print out a warning, as someone should
            # probably take a look!
            if (arg.solver.status == coopr.opt.SolverStatus.warning):
                print('WARNING - Loading a SolverResults object with a '       \
                      'warning status')

            if len(arg.solution) > 0:
                self._load_solution(
                    arg.solution(id),
                    symbol_map=arg.__dict__.get('_symbol_map', None),
                    allow_consistent_values_for_fixed_vars=allow_consistent_values_for_fixed_vars,
                    comparison_tolerance_for_fixed_vars=comparison_tolerance_for_fixed_vars,
                    ignore_invalid_labels=ignore_invalid_labels )
                return True
            else:
                return False
        elif type(arg) is coopr.opt.Solution:
            # set the "stale" flag of each variable in the model prior to loading the
            # solution, so you known which variables have "real" values and which ones don't.
            self.flag_vars_as_stale()
            
            self._load_solution(
                arg,
                symbol_map=symbol_map,
                allow_consistent_values_for_fixed_vars=allow_consistent_values_for_fixed_vars,
                comparison_tolerance_for_fixed_vars=comparison_tolerance_for_fixed_vars,
                ignore_invalid_labels=ignore_invalid_labels )
            return True
        else:
            msg = "Cannot load model with object of type '%s'"
            raise ValueError(msg % str( type(arg) ))

    def Xstore_info(self, results):
        """
        Store model information into a SolverResults object
        """
        results.problem.name = self.name

        stat_keys = (
          'number_of_variables',
          'number_of_binary_variables',
          'number_of_integer_variables',
          'number_of_continuous_variables',
          'number_of_constraints',
          'number_of_objectives'
        )

        for key in stat_keys:
            results.problem.__dict__[key] = self.statistics.__dict__[key]

    def _tuplize(self, data, setobj):
        if data is None:            #pragma:nocover
            return None
        if setobj.dimen == 1:
            return data
        ans = {}
        for key in data:
            if type(data[key][0]) is tuple:
                return data
            ans[key] = tuplize(data[key], setobj.dimen, setobj.name)
        return ans

    def _load_model_data(self, modeldata, namespaces, **kwds):
        """
        Load declarations from a DataPortal object.
        """
        #
        # As we are primarily generating objects here (and acyclic ones
        # at that), there is no need to run the GC until the entire
        # model is created.  Simple reference-counting should be
        # sufficient to keep memory use under control.
        #
        suspend_gc = PauseGC()

        #
        # Unlike the standard method in the pympler summary module, the tracker
        # doesn't print 0-byte entries to pad out the limit.
        #
        profile_memory = kwds.get('profile_memory', 0)

        #
        # It is often useful to report timing results for various activities during model construction.
        #
        report_timing = kwds.get('report_timing', False)

        if (pympler_available is True) and (profile_memory >= 2):
            mem_used = muppy.get_size(muppy.get_objects())
            print("")
            print("      Total memory = %d bytes prior to model construction" % mem_used)

        if (pympler_available is True) and (profile_memory >= 3):
            gc.collect()
            mem_used = muppy.get_size(muppy.get_objects())
            print("      Total memory = %d bytes prior to model construction (after garbage collection)" % mem_used)

        #
        # Do some error checking
        #
        for namespace in namespaces:
            if not namespace is None and not namespace in modeldata._data:
                msg = "Cannot access undefined namespace: '%s'"
                raise IOError(msg % namespace)

        #
        # Initialize each component in order.
        #

        if report_timing is True:
            construction_start_time = time.time()

        for component_name, component in iteritems(self.subcomponents()):

            if component.type() is Model:
                continue

            if report_timing is True:
                start_time = time.time()
                clone_counters = (
                    coopr.pyomo.base.expr.generate_expression.clone_counter,
                    coopr.pyomo.base.expr.generate_relational_expression.clone_counter,
                    coopr.pyomo.base.intrinsic_functions.generate_intrinsic_function_expression.clone_counter,
                    )

            self._initialize_component(modeldata, namespaces, component_name, profile_memory)

            if report_timing is True:
                total_time = time.time() - start_time
                comp = self.find_component(component_name)
                if isinstance(comp, SparseIndexedComponent):
                    clen = len(comp)
                else:
                    assert isinstance(comp, Component)
                    clen = 1
                print("    %%6.%df seconds required to construct component=%s; %d indicies total" \
                          % (total_time>=0.005 and 2 or 0, component_name, clen) \
                          % total_time)
                tmp_clone_counters = (
                    coopr.pyomo.base.expr.generate_expression.clone_counter,
                    coopr.pyomo.base.expr.generate_relational_expression.clone_counter,
                    coopr.pyomo.base.intrinsic_functions.generate_intrinsic_function_expression.clone_counter,
                    )
                if clone_counters != tmp_clone_counters:
                    clone_counters = tmp_clone_counters
                    print("             Cloning detected! (clone counters: %d, %d, %d)" % clone_counters)

        # Note: As is, connectors are expanded when using command-line pyomo but not calling model.create(...) in a Python script.
        # John says this has to do with extension points which are called from commandline but not when writing scripts.
        # Uncommenting the next two lines switches this (command-line fails because it tries to expand connectors twice)
        #connector_expander = ConnectorExpander()
        #connector_expander.apply(instance=self)

        if report_timing is True:
            total_construction_time = time.time() - construction_start_time
            print("      %6.2f seconds required to construct instance=%s" % (total_construction_time, self.name))

        if (pympler_available is True) and (profile_memory >= 2):
            print("")
            print("      Summary of objects following instance construction")
            post_construction_summary = summary.summarize(muppy.get_objects())
            summary.print_(post_construction_summary, limit=100)

            print("")


    def _initialize_component(self, modeldata, namespaces, component_name, profile_memory):
        declaration = self.subcomponent(component_name)

        if component_name in modeldata._default.keys():
            if declaration.type() is not Set:
                declaration.set_default(modeldata._default[component_name])
        data = None
        
        for namespace in namespaces:
            if component_name in modeldata._data.get(namespace,{}).keys():
                if declaration.type() is Set:
                    data = self._tuplize(modeldata._data[namespace][component_name],
                                         declaration)
                else:
                    data = modeldata._data[namespace][component_name]
            if not data is None:
                break

        if __debug__ and logger.isEnabledFor(logging.DEBUG):
            _blockName = "Model" if self.parent() is None \
                else "Block '%s'" % self.cname(True)
            logger.debug( "Constructing %s '%s' on %s from data=%s",
                          declaration.__class__.__name__, 
                          declaration.cname(), _blockName, str(data) )
        try:
            declaration.construct(data)
        except:
            err = sys.exc_info()[1]
            logger.error(
                "Constructing component '%s' from data=%s failed:\n%s: %s",
                str(declaration.cname(True)), str(data).strip(),
                type(err).__name__, err )
            raise

        if __debug__:
            if logger.isEnabledFor(logging.DEBUG):
                _out = StringIO()
                declaration.pprint(ostream=_out)
                logger.debug("Constructed component '%s':\n%s" 
                             % ( declaration.cname(True), _out.getvalue()))
                
        if (pympler_available is True) and (profile_memory >= 2):
            mem_used = muppy.get_size(muppy.get_objects())
            print("      Total memory = %d bytes following construction of component=%s" % (mem_used, component_name))

        if (pympler_available is True) and (profile_memory >= 3):
            gc.collect()
            mem_used = muppy.get_size(muppy.get_objects())
            print("      Total memory = %d bytes following construction of component=%s (after garbage collection)" % (mem_used, component_name))


    def _load_solution( self, soln, symbol_map,
                        allow_consistent_values_for_fixed_vars=False,
                        comparison_tolerance_for_fixed_vars=1e-5,
                        ignore_invalid_labels=False ):
        """
        Load a solution. A solution can be either a tuple or list, or a coopr.opt.Solution instance.
        - The allow_consistent_values_for_fixed_vars flag indicates whether a solution can specify
          consistent values for variables in the model that are fixed.
        - The ignore_invalid_labels flag indicates whether labels in the solution that don't
          appear in the model yield an error. This allows for loading a results object 
          generated from one model into another related, but not identical, model.
        """
        if symbol_map is None:
            same_instance = False
        else:
            same_instance = symbol_map.instance() is self

        # Generate the list of active import suffixes on this top level model
        valid_import_suffixes = dict(active_import_suffix_generator(self))
        # To ensure that import suffix data gets properly overwritten (e.g.,
        # the case where nonzero dual values exist on the suffix and but only
        # sparse dual values exist in the results object) we clear all active
        # import suffixes. 
        for suffix in itervalues(valid_import_suffixes):
            suffix.clearAllValues()

        # Load problem (model) level suffixes. These would only come from ampl
        # interfaced solution suffixes at this point in time.
        for _attr_key, attr_value in iteritems(soln.problem):
            attr_key = _attr_key[0].lower() + _attr_key[1:]
            if attr_key in valid_import_suffixes:
                # GAH: Unlike the var suffix information in the solution object,
                #      problem suffix values are ScalarData objects. I
                #      think it could be advantageous to make all suffix information
                #      ScalarData types. But for now I will take the simple route
                #      and maintain consistency with var suffixes, hence 
                #      attr_value.value rather than just attr_value
                valid_import_suffixes[attr_key].setValue(self,attr_value.value,expand=False)

        #
        # Load objective data (should simply be suffixes if they exist)
        #
        objective_skip_attrs = ['id','canonical_label','value']
        for label,entry in iteritems(soln.objective):

            if same_instance:
                obj_value = symbol_map.getObject(label)
            elif symbol_map is None:
                # We are going to assume the Solution was labeled with
                # the SolverResults pickler
                obj_value = self.find_component(label)
            else:
                obj_value = symbol_map.getEquivalentObject(label, self)

            if obj_value is SymbolMap.UnknownSymbol:
                if ignore_invalid_labels is True:
                    continue
                else:
                    raise KeyError("Objective with label '%s' is not in model '%s'."
                                   % ( label, self.name, ))

            if not isinstance(obj_value, _ObjectiveData):
                raise TypeError("Objective '%s' in model '%s' is type %s"
                                % ( label, self.name, str(type(obj_value)) ))

            for _attr_key, attr_value in iteritems(entry):
                attr_key = _attr_key[0].lower() + _attr_key[1:]
                if attr_key in valid_import_suffixes:
                    # GAH: Unlike the var suffix information in the solution object,
                    #      objective suffix values are ScalarData objects. I
                    #      think it could be advantageous to make all suffix information
                    #      ScalarData types. But for now I will take the simple route
                    #      and maintain consistency with var suffixes, hence 
                    #      attr_value.value rather than just attr_value
                    valid_import_suffixes[attr_key].setValue(obj_value,attr_value.value,expand=False)
        
        #
        # Load variable data
        #
        var_skip_attrs = ['id','canonical_label']
        for label, entry in iteritems(soln.variable):
            if same_instance:
                var_value = symbol_map.getObject(label)
            elif symbol_map is None:
                # We are going to assume the Solution was labeled with
                # the SolverResults pickler
                if 'canonical_label' in entry:
                    var_value = self.find_component(entry['canonical_label'])
                else:
                    # A last-ditch effort to resolve the object by the
                    # old lp-style labeling scheme
                    tmp = label
                    if tmp[-1] == ')': tmp = tmp[:-1]
                    tmp = tmp.replace('(',':').replace(')','.')
                    var_value = self.find_component(tmp)
            else:
                var_value = symbol_map.getEquivalentObject(label, self)

            if var_value is SymbolMap.UnknownSymbol:
                # NOTE: the following is a hack, to handle the ONE_VAR_CONSTANT
                #    variable that is necessary for the objective constant-offset
                #    terms.  probably should create a dummy variable in the model
                #    map at the same time the objective expression is being
                #    constructed.
                if label == "ONE_VAR_CONSTANT":
                    continue
                elif ignore_invalid_labels is True:
                    continue
                else:
                    raise KeyError("Variable label '%s' is not in model '%s'."
                                   % ( label, self.name, ))

            if not isinstance(var_value,_VarData):
                msg = "Variable '%s' in model '%s' is type %s"
                raise TypeError(msg % (
                    label, self.name, str(type(var_value)) ))

            if (allow_consistent_values_for_fixed_vars is False) and (var_value.fixed is True):
                msg = "Variable '%s' in model '%s' is currently fixed - new" \
                      ' value is not expected in solution'
                raise TypeError(msg % ( label, self.name ))

            for _attr_key, attr_value in iteritems(entry):
                attr_key = _attr_key[0].lower() + _attr_key[1:]
                if attr_key == 'value':
                    if (allow_consistent_values_for_fixed_vars is True) and (var_value.fixed is True) and (math.fabs(attr_value - var_value.value) > comparison_tolerance_for_fixed_vars):
                        msg = "Variable '%s' in model '%s' is currently fixed - a value of '%s' in solution is not within tolerance=%s of the current value of '%s'"
                        raise TypeError(msg % ( label, self.name, str(attr_value), str(comparison_tolerance_for_fixed_vars), str(var_value.value) ))
                    var_value.value = attr_value
                    var_value.stale = False
                elif attr_key in valid_import_suffixes:
                    valid_import_suffixes[attr_key].setValue(var_value,attr_value,expand=False)

        #
        # Load constraint data
        #
        con_skip_attrs = ['id', 'canonical_label']
        for label, entry in iteritems(soln.constraint):

            if same_instance:
                con_value = symbol_map.getObject(label)
            elif symbol_map is None:
                # We are going to assume the Solution was labeled with
                # the SolverResults pickler
                con_value = self.find_component(label)
            else:
                con_value = symbol_map.getEquivalentObject(label, self)

            if con_value is SymbolMap.UnknownSymbol:
                #
                # This is a hack - see above.
                #
                if label.endswith('ONE_VAR_CONSTANT'):
                    continue
                elif ignore_invalid_labels is True:
                    continue
                else:
                    raise KeyError("Constraint with label '%s' is not in model '%s'."
                                   % ( label, self.name, ))

            if not isinstance(con_value, _ConstraintData):
                raise TypeError("Constraint '%s' in model '%s' is type %s"
                                % ( label, self.name, str(type(con_value)) ))

            for _attr_key, attr_value in iteritems(entry):
                attr_key = _attr_key[0].lower() + _attr_key[1:]
                if attr_key in valid_import_suffixes:
                    # GAH: Unlike the var suffix information in the solution object,
                    #      constraint suffix values are ScalarData objects. I
                    #      think it could be advantageous to make all suffix information
                    #      ScalarData types. But for now I will take the simple route
                    #      and maintain consistency with var suffixes, hence 
                    #      attr_value.value rather than just attr_value
                    # JPW: The use of MapContainers was nixed for constraints (they have
                    #      long been gone for variables), due to their excessive memory
                    #      requirements. so at the moment, attr_value objects are not
                    #      ScalarData types. These container modifications do not, however,
                    #      contradict Gabe's desire above.
                    valid_import_suffixes[attr_key].setValue(con_value,attr_value,expand=False)
    
    def write(self, filename=None, format=ProblemFormat.cpxlp, solver_capability=None, io_options={}):
        """
        Write the model to a file, with a given format.

        TODO: verify that this method needs to return the filename and symbol_map.
        TODO: these should be returned in a Bunch() object.
        """
        
        if format is None and not filename is None:
            #
            # Guess the format if none is specified
            #
            format = guess_format(filename)
        if solver_capability is None:
            solver_capability = lambda x: True

        problem_writer = coopr.opt.WriterFactory(format)
        if problem_writer is None:
            raise ValueError(\
                    "Cannot write model in format '%s': no model writer " \
                    "registered for that format" \
                    % str(format))

        (filename, symbol_map) = problem_writer(self, filename, solver_capability, io_options)

        if __debug__:
            if logger.isEnabledFor(logging.DEBUG):
                logger.debug("Writing model '%s' to file '%s' with format %s", self.name, str(filename), str(format))
        return filename, symbol_map


class ConcreteModel(Model):
    """
    A concrete optimization model that does not defer construction of
    components.
    """

    def __init__(self, *args, **kwds):
        kwds['_deprecate'] = False
        Model.__init__(self, *args, **kwds)
        self.config.create_functor = 'pyomo.model.default_constructor'
        self.construct()


class AbstractModel(Model):
    """
    An abstract optimization model that defers construction of
    components.
    """

    def __init__(self, *args, **kwds):
        kwds['_deprecate'] = False
        Model.__init__(self, *args, **kwds)
        self.config.create_functor = 'pyomo.model.default_constructor'



@coopr.core.coopr_api(namespace='pyomo.model')
def default_constructor(data, model=None, filename=None, data_dict={}, name=None, namespace=None, namespaces=None, preprocess=True, profile_memory=0, report_timing=False, clone=None):
    """
    Create a concrete instance of this Model, possibly using data
    read in from a file.

    Required:
        model:              An AbstractModel object.

    Optional:
        filename:           The name of a Pyomo Data File that will be used to load
                                data into the model.
        data_dict:          A dictionary containing initialization data for the model
                                to be used if there is no filename
        name:               The name given to the model.
        namespace:          A namespace used to select data.
        namespaces:         A list of namespaces used to select data.
        preprocess:         If False, then preprocessing is suppressed.
        profile_memory:     A number that indicates the profiling level.
        report_timing:      Report timing statistics during construction.
        clone:              Force a clone of the model if this is True.

    Return:
        instance:           Return the model that is constructed.
    """
    if name is None:
        name = model.name
    #
    # Generate a warning if this is a concrete model but the filename is specified.
    # A concrete model is already constructed, so passing in a data file is a waste 
    # of time.
    #
    if model.is_constructed() and isinstance(filename,basestring):
        msg = "The filename=%s will not be loaded - supplied as an argument to the create() method of a ConcreteModel instance with name=%s." % (filename, name)
        logger.warning(msg)
    #
    # If construction is deferred, then clone the model and 
    #    
    if not model._constructed:
        instance = model.clone()

        if namespaces is None or len(namespaces) == 0:
            if filename is None:
                instance.load(data_dict, namespaces=[None], profile_memory=profile_memory, report_timing=report_timing)
            else:
                instance.load(filename, namespaces=[None], profile_memory=profile_memory, report_timing=report_timing)
        else:
            if filename is None:
                instance.load(data_dict, namespaces=namespaces+[None], profile_memory=profile_memory, report_timing=report_timing)
            else:
                instance.load(filename, namespaces=namespaces+[None], profile_memory=profile_memory, report_timing=report_timing)
    else:
        if clone:
            instance = model.clone()
        else:
            instance = model
    #
    # Preprocess the new model
    #    
    if preprocess is True:

        if report_timing is True:
            start_time = time.time()

        instance.preprocess()

        if report_timing is True:
            total_time = time.time() - start_time
            print("      %6.2f seconds required for preprocessing" % total_time)

        if (pympler_available is True) and (profile_memory >= 2):
            mem_used = muppy.get_size(muppy.get_objects())
            print("      Total memory = %d bytes following instance preprocessing" % mem_used)
            print("")

        if (pympler_available is True) and (profile_memory >= 2):
            print("")
            print("      Summary of objects following instance preprocessing")
            post_preprocessing_summary = summary.summarize(muppy.get_objects())
            summary.print_(post_preprocessing_summary, limit=100)

    if not name is None:
        instance.name=name

    return Bunch(instance=instance)


register_component(Model, 'Model objects can be used as a component of other models.')
register_component(ConcreteModel, 'A concrete optimization model that does not defer construction of components.')
register_component(AbstractModel, 'An abstract optimization model that defers construction of components.')




#  _________________________________________________________________________
#
#  Coopr: A COmmon Optimization Python Repository
#  Copyright (c) 2008 Sandia Corporation.
#  This software is distributed under the BSD License.
#  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
#  the U.S. Government retains certain rights in this software.
#  For more information, see the Coopr README.txt file.
#  _________________________________________________________________________

__all__ = ['Component', 'ComponentUID', 'cname']

from weakref import ref as weakref_ref
import sys
from six import iteritems, string_types
import coopr.core
from coopr.pyomo.base.plugin import register_component
from coopr.pyomo.base.misc import tabular_writer


def _cname_index_generator(idx):
    """
    Return a string representation of an index.
    """
    if idx.__class__ is tuple:
        return "[" + ",".join(str(i) for i in idx) + "]"
    else:
        return "[" + str(idx) + "]"


def cname(component, index=None, fully_qualified=False):
    """
    Return a string representation of component for a specific
    index value.
    """
    base = component.cname(fully_qualified=fully_qualified)
    if index is None:
        return base
    else:
        if index not in component.index_set():
            raise KeyError( "Index %s is not valid for component %s"
                            % (index, component.cname(True)) )
        return base + _cname_index_generator( index )


class Component(object):
    """
    This is the base class for all Pyomo modeling components.

    Constructor arguments:
        ctype           The class type for the derived subclass
        doc             A text string describing this component
        name            A name for this component

    Public class attributes:
        doc             A text string describing this component
        name            A name for this component

    Private class attributes:
        _active         A boolean that is true if this component will be 
                            used in model operations
        _constructed    A boolean that is true if this component has been
                            constructed
        _parent         A weakref to the parent block that owns this component
        _type           The class type for the derived subclass
    """

    def __init__ (self, **kwds):
        #
        # Get arguments
        #
        self._type = kwds.pop('ctype', None)
        self.doc   = kwds.pop('doc', None)
        self.name  = kwds.pop('name', str(type(self).__name__))
        if kwds:
            raise ValueError(
                "Unexpected keyword options found while constructing '%s':\n\t%s"
                % ( type(self).__name__, ','.join(sorted(kwds.keys())) ))
        #
        # Verify that ctype has been specified.
        #
        if self._type is None:
            raise coopr.core.DeveloperError("Must specify a class for the component type!")
        #
        self._active        = True
        self._constructed   = False
        self._parent        = None    # Must be a weakref

    @property
    def active(self):
        """Return the active attribute"""
        return self._active

    @active.setter
    def active(self, value):
        """Set the active attribute to the given value"""
        raise AttributeError("Assignment not allowed. Use the (de)activate method")

    def activate(self):
        """Set the active attribute to True"""
        self._active=True

    def deactivate(self):
        """Set the active attribute to False"""
        self._active=False

    def __getstate__(self):
        """
        This method must be defined to support pickling because this class
        owns weakrefs for '_parent'.
        """
        #
        # Nominally, __getstate__() should return:
        #
        # state = super(Class, self).__getstate__()
        # for i in Class.__dict__:
        #     state[i] = getattr(self,i)
        # return state
        #
        # However, in this case, the (nominal) parent class is 'object',
        # and object does not implement __getstate__.  So, we will check
        # to make sure that there is a base __getstate__() to call...
        #
        _base = super(Component,self)
        if hasattr(_base, '__getstate__'):
            state = _base.__getstate__()
            for key,val in iteritems(self.__dict__):
                if key not in state:
                    state[key] = val
        else:
            state = dict(self.__dict__)
        if self._parent is not None:
            state['_parent'] = self._parent()
        return state

    def __setstate__(self, state):
        """
        This method must be defined to support pickling because this class
        owns weakrefs for '_parent'.
        """
        if state['_parent'] is not None and \
                type(state['_parent']) is not weakref_ref:
            state['_parent'] = weakref_ref(state['_parent'])
        #
        # Note: our model for setstate is for derived classes to modify
        # the state dictionary as control passes up the inheritance
        # hierarchy (using super() calls).  All assignment of state ->
        # object attributes is handled at the last class before 'object'
        # (which may -- or may not (thanks to MRO) -- be here.
        #
        _base = super(Component,self)
        if hasattr(_base, '__setstate__'):
            return _base.__setstate__(state)
        else:
            for key, val in iteritems(state):
                # Note: per the Python data model docs, we explicitly
                # set the attribute using object.__setattr__() instead
                # of setting self.__dict__[key] = val.
                object.__setattr__(self, key, val)

    def type(self):
        """Return the class type for this component"""
        return self._type

    def construct(self, data=None):                     #pragma:nocover
        """API definition for constructing components"""
        pass

    def is_constructed(self):                           #pragma:nocover
        """Return True if this class has been constructed"""
        return self._constructed

    def reconstruct(self, data=None):
        """Re-construct model expressions"""
        self._constructed = False
        self.construct(data=data)

    def valid_model_component(self):
        """Return True if this can be used as a model component."""
        return True

    def pprint(self, ostream=None, verbose=False, prefix=""):
        """Print component information"""
        if ostream is None:
            ostream = sys.stdout
        tab="    "
        ostream.write(prefix+self.cname()+" : ")
        if self.doc is not None:
            ostream.write(self.doc+'\n'+prefix+tab)

        _attr, _data, _header, _fcn = self._pprint()

        ostream.write(", ".join("%s=%s" % (k,v) for k,v in _attr))
        ostream.write("\n")
        if not self._constructed:
            ostream.write(prefix+tab+"Not constructed\n")
            return

        if _data is not None:
            tabular_writer( ostream, prefix+tab, _data, _header, _fcn )

    def display(self, ostream=None, verbose=False, prefix=""):
        self.pprint(ostream=ostream, prefix=prefix)

    def component(self):
        """Returns the component associated with this object."""
        return self

    def parent(self):
        """Returns the parent of this object."""
        if self._parent is None:
            return None
        else:
            return self._parent()

    def reset(self):
        """Reset the state of this component."""
        pass

    def model(self):
        """Returns the model associated with this object."""
        ans = self.parent()
        if ans is None:
            return None
        #
        # NOTE: This loop is probably OK, since most models won't be
        # nested very deep.
        #
        while ans.parent() is not None:
            ans = ans.parent()
        return ans

    def __str__(self):
        """Return the component name"""
        return self.cname(True)

    def to_string(self, ostream=None, verbose=None, precedence=0):
        """Write the component name to a buffer"""
        if ostream is None:
            ostream = sys.stdout
        ostream.write(self.__str__())

    def cname(self, fully_qualified=False, name_buffer=None):
        """
        Returns the component name associated with this object.

        Arguments:
            fully_qualified     Generate full name from nested block names
            name_buffer         Can be used to optimize iterative name 
                                    generation (using a dictionary)
        """
        if fully_qualified and self.parent() and self.parent() is not self.model():
            return self.parent().cname(fully_qualified, name_buffer) \
                + "." + self.name
        return self.name

    def pprint(self, ostream=None, verbose=False, prefix=""):
        """Print component information"""
        if ostream is None:
            ostream = sys.stdout
        tab="    "
        ostream.write(prefix+self.cname()+" : ")
        if self.doc is not None:
            ostream.write(self.doc+'\n'+prefix+tab)

        _attr, _data, _header, _fcn = self._pprint()

        ostream.write(", ".join("%s=%s" % (k,v) for k,v in _attr))
        ostream.write("\n")
        if not self._constructed:
            ostream.write(prefix+tab+"Not constructed\n")
            return

        if _data is not None:
            tabular_writer( ostream, prefix+tab, _data, _header, _fcn )


class ComponentData(object):
    """
    This is the base class for the component data used
    in Pyomo modeling components.  Subclasses of ComponentData are
    used in indexed components, and this class assumes that indexed
    components are subclasses of SparseIndexedComponent.  Note that
    ComponentData instances do not store their index.  This makes
    some operations significantly more expensive, but these are (a)
    associated with I/O generation and (b) this cost can be managed 
    with caches.

    Constructor arguments:
        owner           The component that owns this data object

    Private class attributes:
        _component      A weakref to the component that owns this data object
    """

    __slots__ = ( '_component', )

    def __init__(self, owner):
        #
        # ComponentData objects are typically *private* objects for
        # indexed / sparse indexed components.  As such, the (derived)
        # class needs to make sure that the owning component is *always*
        # passed as the owner (and that owner is never None).  Not validating
        # this assumption is significantly faster.
        #
        self._component = weakref_ref(owner)

    def __getstate__(self):
        """
        This method must be defined to support pickling because this class
        owns weakrefs for '_component'.
        """
        #
        # Nominally, __getstate__() should return:
        #
        # state = super(Class, self).__getstate__()
        # for i in Class.__slots__:
        #    state[i] = getattr(self,i)
        # return state
        #
        # However, in this case, the (nominal) parent class is 'object',
        # and object does not implement __getstate__.  So, we will check
        # to make sure that there is a base __getstate__() to call...
        #
        # Further, since there is only a single slot, and that slot
        # (_component) requires special processing, we will just deal
        # with it explicitly.  As _component is a weakref (not
        # pickable), we need to resolve it to a concrete object.
        #    
        _base = super(ComponentData,self)
        if hasattr(_base, '__getstate__'):
            state = _base.__getstate__()
        else:
            state = {}
        #
        if self._component is None:
            state['_component'] = None
        else:
            state['_component'] = self._component()
        return state

    def __setstate__(self, state):
        """
        This method must be defined to support unpickling because this class
        owns weakrefs for '_component'.
        """
        #
        # FIXME: We shouldn't have to check for weakref.ref here, but if
        # we don't the model cloning appears to fail (in the Benders
        # example)
        #
        if state['_component'] is not None and \
                type(state['_component']) is not weakref_ref:
            state['_component'] = weakref_ref(state['_component'])
        #
        # Note: our model for setstate is for derived classes to modify
        # the state dictionary as control passes up the inheritance
        # hierarchy (using super() calls).  All assignment of state ->
        # object attributes is handled at the last class before 'object'
        # (which may -- or may not (thanks to MRO) -- be here.
        #
        _base = super(ComponentData,self)
        if hasattr(_base, '__setstate__'):
            return _base.__setstate__(state)
        else:
            for key, val in iteritems(state):
                # Note: per the Python data model docs, we explicitly
                # set the attribute using object.__setattr__() instead
                # of setting self.__dict__[key] = val.
                object.__setattr__(self, key, val)

    def component(self):
        """Returns the component associated with this object."""
        if self._component is None: 
            return None
        return self._component()

    def parent(self):
        """Return the parent of the component that owns this data. """
        ans = self.component()
        if ans is None:
            return None
        #
        # Directly call the Component's model() to prevent infinite
        # recursion for scalar component.
        #
        if ans is self:
            return super(ComponentData, ans).parent()
        else:
            return ans.parent()

    def model(self):
        """Return the model of the component that owns this data. """
        ans = self.component()
        if ans is None:
            return None
        #
        # Directly call the Component's model() to prevent infinite
        # recursion for scalar objects.
        #
        if ans is self:
            return super(ComponentData, ans).model()
        else:
            return ans.model()

    def index(self):
        """
        Returns the index of this ComponentData instance relative
        to the parent component index set. None is returned if 
        this instance does not have a parent component, or if
        - for some unknown reason - this instance does not belong
        to the parent component's index set. This method is not 
        intended to be a fast method;  it should be used rarely, 
        primarily in cases of label formulation.
        """
        self_component = self.component()
        if self_component is None:
            return None
        for idx, component_data in self_component.iteritems():
            if id(component_data) == id(self):
                return idx
        return None
        
    def __str__(self):
        """Return a string with the component name and index"""
        return self.cname(True)

    def to_string(self, ostream=None, verbose=None, precedence=0):
        """Write the component name and index to a buffer"""
        if ostream is None:
            ostream = sys.stdout
        ostream.write(self.__str__())

    def cname(self, fully_qualified=False, name_buffer=None):
        """Return a string with the component name and index"""
        c = self.component()
        if c is self:
            # This is a scalar component, so call the Component.cname() method
            return super(ComponentData, self).cname(fully_qualified, name_buffer)
        #
        # Get the name of the component
        #
        base = c.cname(fully_qualified, name_buffer)
        if name_buffer is not None:
            #
            # Using the buffer, which is a dictionary:  id -> string
            #
            if id(self) in name_buffer:
                # Return the name if it is in the buffer
                return name_buffer[id(self)]
            # Iterate through the dictionary and generate all names in the buffer
            for idx, obj in iteritems(c._data):
                name_buffer[id(obj)] = base + _cname_index_generator(idx)
            if id(self) in name_buffer:
                # Return the name if it is in the buffer
                return name_buffer[id(self)]
        else:
            #
            # No buffer, so we iterate through the component _data 
            # dictionary until we find this object.  This can be much
            # more expensive than if a buffer is provided.
            #
            for idx, obj in iteritems(c._data):
                if obj is self:
                    return base + _cname_index_generator(idx)
        #
        raise RuntimeError("Fatal error: cannot find the component data in "
                           "the owning component's _data dictionary.")


class ActiveComponentData(ComponentData):
    """
    This is the base class for the component data used
    in Pyomo modeling components that can be activated and
    deactivated.

    It's possible to end up in a state where the parent Component
    has _active=True but all ComponentData have _active=False. This
    seems like a reasonable state, though we cannot easily detect
    this situation.  The important thing to avoid is the situation
    where one or more ComponentData are active, but the parent
    Component claims active=False. This class structure is designed
    to prevent this situation.

    Constructor arguments:
        owner           The component that owns this data object

    Private class attributes:
        _component      A weakref to the component that owns this data object
        _active         A boolean that indicates whether this data is active
    """

    __slots__ = ( '_active', )

    def __init__(self, owner):
        super(ActiveComponentData, self).__init__(owner)
        self._active = True

    def __getstate__(self):
        """
        This method must be defined because this class uses slots.
        """
        result = super(ActiveComponentData, self).__getstate__()
        for i in ActiveComponentData.__slots__:
            result[i] = getattr(self, i)
        return result

    # Since this class requires no special processing of the state
    # dictionary, it does not need to implement __setstate__()

    @property
    def active(self):
        """Return the active attribute"""
        return self._active

    @active.setter
    def active(self, value):
        """Set the active attribute to a specified value."""
        raise AttributeError("Assignment not allowed. Use the (de)activate method")

    def activate(self):
        """Set the active attribute to True"""
        self._active = self.component()._active = True

    def deactivate(self):
        """Set the active attribute to False"""
        self._active = False


class ComponentUID(object):
    """This class provides a system to generate "component unique
    identifiers".  Any component in a model can be described by a CUID,
    and from a CUID you can find the component.  An important feature of
    CUIDs is that they are relative to a model, so you can use a CUID
    generated on one model to find the equivalent component on another
    model.  This is especially useful when you clone a model and want
    to, for example, copy a variable value from the cloned model back to
    the original model.

    The CUID has a string representation that can specify a specific
    component or a group of related components through the use of index
    wildcards (* for a single element in the index, and ** for all
    indexes)

    This class is also used by test_component.py to validate the structure
    of components.

    """

    __slots__ = ( '_cids', )
    tList = [ int, str ]
    tKeys = '#$'
    tDict = {} # ...initialized below

    def __init__(self, component, cuid_buffer=None):
        # A CUID can be initialized from either a reference component or
        # the string representation.
        if isinstance(component, string_types):
            self._cids = tuple( self.parse_cuid(component) )
        else:
            self._cids = tuple( self._generate_cuid(component, cuid_buffer) )

    def __str__(self):
        a = ""
        for name, args, types in reversed(self._cids):
            if a:
                a += '.' + name
            else:
                a = name
            if types is None:
                a += '[**]'
                continue
            if len(args) == 0:
                continue
            a += '['+','.join(str(x) or '*' for x in args) + ']'
        return a

    def __repr__(self):
        a = ""
        for name, args, types in reversed(self._cids):
            if a:
                a += '.' + name
            else:
                a = name
            if types is None:
                a += ':**'
                continue
            if len(args) == 0:
                continue
            a += ':'+','.join( (types[i] if types[i] not in '.' else '')+str(x) 
                               for i,x in enumerate(args) )
        return a

    def __getstate__(self):
        return dict((x,getattr(val,x)) for x in ComponentUID.__slots__)

    def __setstate__(self, state):
        for key, val in iteritems(state):
            setattr(self,key,val) 

    # Define all comparison operators using the underlying tuple's
    # comparison operators. We will be lazy and assume that the other is
    # a CUID.

    def __hash__(self):
        return self._cids.__hash__()

    def __lt__(self, other):
        try:
            return self._cids.__lt__(other._cids)
        except AttributeError:
            return self._cids.__lt__(other)

    def __le__(self, other):
        try:
            return self._cids.__le__(other._cids)
        except AttributeError:
            return self._cids.__le__(other)

    def __gt__(self, other):
        try:
            return self._cids.__gt__(other._cids)
        except AttributeError:
            return self._cids.__gt__(other)

    def __ge__(self, other):
        try:
            return self._cids.__ge__(other._cids)
        except AttributeError:
            return self._cids.__ge__(other)

    def __eq__(self, other):
        try:
            return self._cids.__eq__(other._cids)
        except AttributeError:
            return self._cids.__eq__(other)

    def __ne__(self, other):
        try:
            return self._cids.__ne__(other._cids)
        except AttributeError:
            return self._cids.__ne__(other)

    def _partial_cuid_from_index(self, idx):
        tDict = ComponentUID.tDict
        if idx.__class__ is tuple:
            return ( idx, ''.join(tDict.get(type(x), '?') for x in idx) )
        else:
            return ( (idx,), tDict.get(type(idx), '?') )

    def _generate_cuid(self, component, cuid_buffer=None):
        model = component.model()
        tDict = ComponentUID.tDict
        if not hasattr(component, '_component'):
            yield ( component.cname(), '**', None )
            component = component.parent()
        while component is not model:
            c = component.component()
            if c is component:
                yield ( c.cname(), tuple(), '' )
            elif cuid_buffer is not None:
                if id(self) not in cuid_buffer:
                    for idx, obj in iteritems(c._data):
                        if idx.__class__ is tuple:
                            cuid_buffer[id(obj)] = \
                                self._partial_cuid_from_index(idx)
                yield (c.cname(),) + cuid_buffer[id(self)]
            else:
                for idx, obj in iteritems(c._data):
                    if obj is component:
                        yield (c.cname(),) + self._partial_cuid_from_index(idx)
                        break
            component = component.parent()

    def parse_cuid(self, label):
        cList = label.split('.')
        tKeys = ComponentUID.tKeys
        tDict = ComponentUID.tDict
        for c in reversed(cList):
            if c[-1] == ']':
                c_info = c[:-1].split('[',1)
            else:
                c_info = c.split(':',1)
            if len(c_info) == 1:
                yield ( c_info[0], tuple(), '' )
            else:
                idx = c_info[1].split(',')
                _type = ''
                for i, val in enumerate(idx):
                    if val == '*':
                        _type += '*'
                        idx[i] = ''
                    elif val[0] in tKeys:
                        _type += val[0]
                        idx[i] = tDict[val[0]](val[1:])
                    elif val[0] in  "\"'" and val[-1] == val[0]:
                        _type += ComponentUID.tDict[str]
                        idx[i] = val[1:-1]
                    else:
                        _type += '.'
                if len(idx) == 1 and idx[0] == '**':
                    yield ( c_info[0], '**', None )
                else:
                    yield ( c_info[0], tuple(idx), _type )

    def find_component_on(self, model):
        return self.find_component(model)

    # Return the (unique) component in the model.  If the CUID contains
    # a wildcard in the last component, then returns that component.  If
    # there are wildcards elsewhere (or the last component was a partial
    # slice), then returns None.  See list_components below.
    def find_component(self, model):
        obj = model
        for name, idx, types in reversed(self._cids):
            try:
                if len(idx) and idx != '**' and types.strip('*'):
                    obj = getattr(obj, name)[idx]
                else:
                    obj = getattr(obj, name)
            except KeyError:
                if '.' not in types:
                    return None
                tList = ComponentUID.tList
                def _checkIntArgs(_idx, _t, _i):
                    if _i == -1:
                        try:
                            return getattr(obj, name)[tuple(_idx)]
                        except KeyError:
                            return None
                    _orig = _idx[_i]
                    for _cast in tList:
                        try:
                            _idx[_i] = _cast(_orig)
                            ans = _checkIntArgs(_idx, _t, _t.find('.',_i+1))
                            if ans is not None:
                                return ans
                        except ValueError:
                            pass
                    _idx[_i] = _orig
                    return None
                obj = _checkIntArgs(list(idx), types, types.find('.'))
            except AttributeError:
                return None
        return obj

    def _list_components(self, _obj, cids):
        if not cids:
            yield _obj
            return

        name, idx, types = cids[-1]
        try:
            obj = getattr(_obj, name)
        except AttributeError:
            return
        if len(idx) == 0:
            for ans in self._list_components(obj, cids[:-1]):
                yield ans
        elif idx != '**' and '*' not in types and '.' not in types:
            try:
                obj = obj[idx]
            except KeyError:
                return
            for ans in self._list_components(obj, cids[:-1]):
                yield ans
        else:
            all =  idx == '**'
            tList = ComponentUID.tList
            for target_idx, target_obj in iteritems(obj._data):
                if not all and idx != target_idx:
                    _idx, _types = self._partial_cuid_from_index(target_idx)
                    if len(idx) != len(_idx):
                        continue
                    match = True
                    for j in range(len(idx)):
                        if idx[j] == _idx[j] or types[j] == '*':
                            continue
                        elif types[j] == '.':
                            ok = False
                            for _cast in tList:
                                try:
                                    if _cast(idx[j]) == _idx[j]:
                                        ok = True
                                        break
                                except ValueError:
                                    pass
                            if not ok:
                                match = False
                                break
                        else:
                            match = False
                            break
                    if not match:
                        continue
                for ans in self._list_components(target_obj, cids[:-1]):
                    yield ans

    def list_components(self, model):
        for ans in self._list_components(model, self._cids):
            yield ans

    def matches(self, component):
        tList = ComponentUID.tList
        for i, (name, idx, types) in enumerate(self._generate_cuid(component)):
            if i == len(self._cids):
                return False
            _n, _idx, _types = self._cids[i]
            if _n != name:
                return False
            if _idx == '**' or idx == _idx:
                continue
            if len(idx) != len(_idx):
                return False
            for j in range(len(idx)):
                if idx[j] == _idx[j] or _types[j] == '*':
                    continue
                elif _types[j] == '.':
                    ok = False
                    for _cast in tList:
                        try:
                            if _cast(_idx[j]) == idx[j]:
                                ok = True
                                break
                        except ValueError:
                            pass
                    if not ok:
                        return False
                else:
                    return False
        # Matched if all self._cids were consumed
        return i+1 == len(self._cids)

# WEH - What does it mean to initialize this dictionary outside 
#       of the definition of this class?  Is tList populated
#       with all components???
ComponentUID.tDict.update( (ComponentUID.tKeys[i], v) 
                           for i,v in enumerate(ComponentUID.tList) )
ComponentUID.tDict.update( (v, ComponentUID.tKeys[i]) 
                           for i,v in enumerate(ComponentUID.tList) )






from coopr.pyomo import *
from sc import *

@pyomo_callback('solve-callback')
def solve_callback(solver, model):
    print "CB-Solve"

@pyomo_callback('cut-callback')
def cut_callback(solver, model):
    print "CB-Cut"

@pyomo_callback('node-callback')
def node_callback(solver, model):
    print "CB-Node"





import ast
import re

from coopr.pyomo.plugins.check.model import ModelTrackerHook
from coopr.pyomo.plugins.check.checker import IterativeTreeChecker


class ModelName(IterativeTreeChecker):

    def beginChecking(self, runner, script):
        self.modelAssigned = False

    def endChecking(self, runner, script):
        if not self.modelAssigned:
            self.problem("Global object 'model' is never assigned.")

    def checkerDoc(self):
        return """\
        Pyomo will be unable to execute this model
        file without additional arguments.
        """

    def checkTarget(self, node):
        if isinstance(node, ast.Name) and node.id == 'model':
            self.modelAssigned = True

    def check(self, runner, script, info):
        # If assigning, check target name
        if isinstance(info, ast.Assign):
            for target in info.targets:
                self.checkTarget(target)

                # Handle multiple assignment
                if isinstance(target, ast.Tuple):
                    for elt in target.elts:
                        self.checkTarget(elt)


class ModelCreate(IterativeTreeChecker):

    def getTargetStrings(self, assign):
        ls = []
        for target in assign.targets:
            if isinstance(target, ast.Name):
                ls.append(target.id)
            elif isinstance(target, ast.Tuple):
                ls.extend(list(map((lambda x: x.id), target.elts))) # TODO probably not resilient
        return ls

    def checkerDoc(self):
        return """\
        Usually, developers create an instance of a Pyomo model class.
        It is rare that the class itself needs to be assigned to a
        variable, e.g.:
            x = ConcreteModel      =>      x = ConcreteModel()
        """

    def check(self, runner, script, info):
        if isinstance(info, ast.Assign):
            if 'model' in self.getTargetStrings(info):
                if isinstance(info.value, ast.Name):
                    if info.value.id.endswith("Model"):
                        self.problem("Possible incorrect assignment of " + info.value.id + " class instead of instance", lineno = info.lineno)


class DeprecatedModel(IterativeTreeChecker):

    def checkerDoc(self):
        return """\
        The Model class is no longer supported as an object that can be
        created - instead, use the ConcreteModel or AbstractModel class.
        """

    def check(self, runner, script, info):
        if isinstance(info, ast.Assign):
            if isinstance(info.value, ast.Call):
                if isinstance(info.value.func, ast.Name):
                    if info.value.func.id == 'Model':
                        self.problem("Deprecated use of Model class", lineno = info.value.func.lineno)



#
# Test the canonical expressions
#

import os
import sys
from os.path import abspath, dirname
currdir = dirname(abspath(__file__))+os.sep

from six import iteritems
import pyutilib.th as unittest
import pyutilib.services
from coopr.pyomo import *
from coopr.pyomo.base.expr import Expr_if
from coopr.opt import *

class frozendict(dict):
    __slots__ = ('_hash',)
    def __hash__(self):
        rval = getattr(self, '_hash', None)
        if rval is None:
            rval = self._hash = hash(frozenset(iteritems(self)))
        return rval

# a silly utility to facilitate comparison of tuples where we don't care about ordering
def linear_repn_to_dict(repn):
    result = {}
    if repn.variables is not None:
        for i in xrange(len(repn.variables)):
            result[id(repn.variables[i])] = repn.linear[i]
    if repn.constant != None:
        result[None] = repn.constant
    return result

class Test(unittest.TestCase):

    #def setUp(self):
        #
        # Create Model
        #
        #self.plugin = SimplePreprocessor()
        #self.plugin.deactivate_action("compute_canonical_repn")

    def tearDown(self):
        if os.path.exists("unknown.lp"):
            os.unlink("unknown.lp")
        pyutilib.services.TempfileManager.clear_tempfiles()
        #self.plugin.activate_action("compute_canonical_repn")

    def test_abstract_linear_expression(self):
        m = AbstractModel()
        m.A = RangeSet(1,3)
        def p_init(model, i):
            return 2*i
        m.p = Param(m.A, initialize=p_init)
        m.x = Var(m.A, bounds=(-1,1))
        def obj_rule(model):
            return summation(model.p, model.x)
        m.obj = Objective(rule=obj_rule)
        i = m.create()

        rep = generate_canonical_repn(i.obj[None].expr)
        # rep should only have variables and linear terms
        self.assertTrue(rep.linear != None)
        self.assertTrue(rep.constant == None)
        self.assertTrue(rep.variables != None)
        # rep.variables should have the 3 variables...
        baseline = { id(i.x[1]) : 2, 
                     id(i.x[2]) : 4,
                     id(i.x[3]) : 6 }
        self.assertEqual(baseline,
                         linear_repn_to_dict(rep))

    def test_concrete_linear_expression(self):
        m = ConcreteModel()
        m.A = RangeSet(1,3)
        def p_init(model, i):
            return 2*i
        m.p = Param(m.A, initialize=p_init)
        m.x = Var(m.A, bounds=(-1,1))
        m.obj = Objective(expr=summation(m.p, m.x))
        rep = generate_canonical_repn(m.obj[None].expr)
        # rep should only have variables and linear terms
        self.assertTrue(rep.variables != None)
        self.assertTrue(rep.linear != None)
        # rep.variables should have the 3 variables...
        baseline = { id(m.x[1]) : 2, 
                     id(m.x[2]) : 4,
                     id(m.x[3]) : 6 }        
        self.assertEqual(baseline,
                         linear_repn_to_dict(rep))

    def test_linear_expression(self):
        I = range(3)
        x = [Var(bounds=(-1,1)) for i in I]
        expr = sum(2*(i+1)*x[i] for i in I)
        rep = generate_canonical_repn(expr)
        # rep should only have variables and linear terms
        self.assertTrue(rep.variables != None)
        self.assertTrue(rep.linear != None)
        # rep.variables should have the 3 variables... 
        baseline = { id(x[0]) : 2, 
                     id(x[1]) : 4,
                     id(x[2]) : 6 }        
        self.assertEqual(baseline,
                         linear_repn_to_dict(rep))

    def test_complex_linear_expression(self):
        I = range(3)
        x = [Var(bounds=(-1,1)) for i in I]
        expr = 2*x[1] + 3*x[1] + 4*(x[2]+x[1])
        rep = generate_canonical_repn(expr)
        # rep should only have variables and linear components
        self.assertTrue(rep.linear != None)
        self.assertTrue(rep.variables != None)
        # rep.variables should have the 2 of the 3 variables...
        baseline = { id(x[1]) : 9.0, 
                     id(x[2]) : 4.0 }
        self.assertEqual(baseline,
                         linear_repn_to_dict(rep))


    def test_reversed_complex_linear_expression(self):
        I = range(3)
        x = [Var(bounds=(-1,1)) for i in I]
        expr = 4*(x[2]+x[1]) + 2*x[1] + 3*x[1]
        rep = generate_canonical_repn(expr)
        # rep should only have variables and linear terms
        self.assertTrue(rep.linear != None)
        self.assertTrue(rep.variables != None)
        # rep.variables should have the 2 of the 3 variables...
        baseline = { id(x[2]) : 4.0, 
                     id(x[1]) : 9.0 }
        self.assertEqual(baseline,
                         linear_repn_to_dict(rep))

    def test_linear_expression_with_constant(self):
        I = range(3)
        x = [Var(bounds=(-1,1)) for i in I]
        expr = 1.2 + 2*x[1] + 3*x[1]
        rep = generate_canonical_repn(expr)
        # rep should only have [-1,0,1]
        self.assertTrue(rep.constant != None)
        self.assertTrue(rep.linear != None)
        self.assertTrue(rep.variables != None)
        # rep should have the 1 of the 3 variables...
        baseline = { id(x[1]) : 5.0,
                     None : 1.2 } 
        self.assertEqual(baseline,
                         linear_repn_to_dict(rep))

    def test_complex_linear_expression_with_constant(self):
        I = range(3)
        x = [Var(bounds=(-1,1)) for i in I]
        expr =  2.0*(1.2 + 2*x[1] + 3*x[1]) + 3.0*(1.0+x[1])
        rep = generate_canonical_repn(expr)
        # rep should only have variables, a constant term, and linear terms
        self.assertTrue(rep.constant != None)
        self.assertTrue(rep.linear != None)
        self.assertTrue(rep.variables != None)
        # rep.variables should have the 1 of the 3 variables...
        baseline = { id(x[1]) : 13.0,
                     None : 5.4 } 
        self.assertEqual(baseline,
                         linear_repn_to_dict(rep))

    def test_polynomial_expression(self):
        I = range(4)
        x = [Var(bounds=(-1,1)) for i in I]
        expr = x[1]*(x[1]+x[2]) + x[2]*(x[1]+3.0*x[3]*x[3])
        rep = generate_canonical_repn(expr)
        # rep should only have [-1,2,3]
        self.assertEqual(len(rep), 3)
        self.assertTrue(2 in rep)
        self.assertTrue(3 in rep)
        self.assertTrue(-1 in rep)
        # rep[-1] should have the 3 of the 4 variables...
        self.assertEqual(rep[-1], { 0: x[1],
                                    1: x[2],
                                    2: x[3] })
        # check the expression encoding
        self.assertEqual(rep[2], {frozendict({0:2}):1.0, 
                                  frozendict({0:1, 1:1}):2.0})
        self.assertEqual(rep[3], {frozendict({1:1, 2:2}):3.0})


    def test_polynomial_expression_with_fixed(self):
        I = range(4)
        x = [Var() for i in I]
        for v in x:
            v.construct()
        expr = x[1]*(x[1]+x[2]) + x[2]*(x[1]+3.0*x[3]*x[3])
        x[1].value = 5
        x[1].fixed = True
        rep = generate_canonical_repn(expr)
        # rep should only have [-1,0,1,3]
        self.assertEqual(len(rep), 4)
        self.assertTrue(0 in rep)
        self.assertTrue(1 in rep)
        self.assertTrue(3 in rep)
        self.assertTrue(-1 in rep)
        # rep[-1] should have the 2 of the 4 variables...
        self.assertEqual(rep[-1], { 0: x[2],
                                    1: x[3] })
        # check the expression encoding
        self.assertEqual(rep[0], {None: 25.0})
        self.assertEqual(rep[1], {0: 10.0})
        self.assertEqual(rep[3], {frozendict({0:1, 1:2}):3.0})


    def test_linear_expression_with_constant_division(self):
        m = ConcreteModel()
        m.A = RangeSet(1,3)
        def p_init(model, i):
            return 2*i
        m.p = Param(m.A, initialize=p_init)
        m.x = Var(m.A, bounds=(-1,1))
        expr = summation(m.p, m.x)/2.0
        rep = generate_canonical_repn(expr)

        # rep should only have only variables and linear terms
        self.assertTrue(rep.linear != None)
        self.assertTrue(rep.constant == None)
        self.assertTrue(rep.variables != None)

        baseline = { id(m.x[1]) : 1.0, 
                     id(m.x[2]) : 2.0,
                     id(m.x[3]) : 3.0 }
        self.assertEqual(baseline,
                         linear_repn_to_dict(rep))

    def test_linear_expression_with_fixed_division(self):
        m = ConcreteModel()
        m.A = RangeSet(1,3)
        def p_init(model, i):
            return 2*i
        m.p = Param(m.A, initialize=p_init)
        m.x = Var(m.A, bounds=(-1,1))
        m.y = Var(initialize=2.0)
        m.y.fixed = True
        expr = summation(m.p, m.x)/m.y
        rep = generate_canonical_repn(expr)

        # rep should only have variables and linearm terms
        self.assertTrue(rep.linear != None)
        self.assertTrue(rep.constant == None)
        self.assertTrue(rep.variables != None)

        # rep.variables should have the 3 variables... 
        baseline = { id(m.x[1]) : 1.0, 
                     id(m.x[2]) : 2.0,
                     id(m.x[3]) : 3.0 }
        self.assertEqual(baseline,
                         linear_repn_to_dict(rep))

    def test_linear_expression_with_complex_fixed_division(self):
        m = ConcreteModel()
        m.A = RangeSet(1,3)
        def p_init(model, i):
            return 2*i
        m.p = Param(m.A, initialize=p_init)
        m.x = Var(m.A, bounds=(-1,1))
        m.y = Var(initialize=1.0)
        m.y.fixed = True
        expr = summation(m.p, m.x)/(m.y+1)
        rep = generate_canonical_repn(expr)

        # rep should only have variable and a linear component
        self.assertTrue(rep.linear != None)
        self.assertTrue(rep.constant == None)
        self.assertTrue(rep.variables != None)

        # rep.variables should have the 3 variables... 
        baseline = { id(m.x[1]) : 1.0, 
                     id(m.x[2]) : 2.0,
                     id(m.x[3]) : 3.0 }
        self.assertEqual(baseline,
                         linear_repn_to_dict(rep))

    def test_expr_rational_summation(self):
        m = ConcreteModel()
        m.A = RangeSet(1,3)
        def p_init(model, i):
            return 2*i
        m.p = Param(m.A, initialize=p_init)
        m.x = Var(m.A, bounds=(-1,1))
        m.y = Var(initialize=1.0)
        expr = summation(m.p, m.x)/(1+m.y)

        rep = generate_canonical_repn(expr)
        # rep should only have [-1,None]
        self.assertEqual(len(rep), 2)
        self.assertTrue(None in rep)
        self.assertTrue(-1 in rep)
        # rep[-1] should have the 4 variables... 
        self.assertEqual(rep[-1], { 3: m.y,
                                    0: m.x[1],
                                    1: m.x[2],
                                    2: m.x[3] })
        # check the expression encoding
        self.assertIs(rep[None], expr)

    def test_expr_rational(self):
        m = ConcreteModel()
        m.A = RangeSet(1,4)
        m.x = Var(m.A, bounds=(-1,1))
        m.y = Var(initialize=1.0)
        expr = (1.25+m.x[1]+1/m.y) + (2.0+m.x[2])/m.y

        rep = generate_canonical_repn(expr)
        # rep should only have [-1,None]
        self.assertEqual(len(rep), 2)
        self.assertTrue(None in rep)
        self.assertTrue(-1 in rep)
        # rep[-1] should have the 3 variables... 
        self.assertEqual(rep[-1], { 1: m.y,
                                    0: m.x[1],
                                    2: m.x[2] })
        # check the expression encoding
        self.assertIs(rep[None], expr)

    def test_expr_rational_fixed(self):
        m = ConcreteModel()
        m.A = RangeSet(1,4)
        m.x = Var(m.A, bounds=(-1,1))
        m.y = Var(initialize=2.0)
        m.y.fixed = True
        expr = (1.25+m.x[1]+1/m.y) + (2.0+m.x[2])/m.y

        rep = generate_canonical_repn(expr)

        # rep should only have variables, a constant, and linear terms
        self.assertTrue(rep.constant != None)
        self.assertTrue(rep.linear != None)
        self.assertTrue(rep.variables != None)

        # rep.variables should have the 2 of 3 variables... 

        baseline = { id(m.x[1]) : 1.0, 
                     id(m.x[2]) : 0.5,
                     None : 2.75 }
        self.assertEqual(baseline,
                         linear_repn_to_dict(rep))

    def test_general_nonlinear(self):
        I = range(3)
        x = [Var() for i in I]
        expr = x[1] + 5*cos(x[2])

        rep = generate_canonical_repn(expr)
        # rep should only have [-1,None]
        self.assertEqual(len(rep), 2)
        self.assertTrue(None in rep)
        self.assertTrue(-1 in rep)
        # rep[-1] should have 2 variables... 
        self.assertEqual(rep[-1], { 0: x[1],
                                    1: x[2] })
        # check the expression encoding
        self.assertIs(rep[None], expr)
        
    def test_general_nonlinear_fixed(self):
        I = range(3)
        x = [Var() for i in I]
        for v in x:
            v.construct()
        expr = x[1] + 5*cos(x[2])
        x[2].value = 0
        x[2].fixed = True

        rep = generate_canonical_repn(expr)
        # rep should only have variables, a constsant, and linear terms
        self.assertTrue(isinstance(rep, LinearCanonicalRepn) == True)
        self.assertTrue(rep.variables != None)
        self.assertTrue(rep.linear != None)
        self.assertTrue(rep.constant != None)

        # rep.variables should have 1 variable... 
        baseline = { id(x[1]) : 1.0,
                     None: 5.0 }
        self.assertEqual(baseline,
                         linear_repn_to_dict(rep))

    def test_deterministic_var_labeling(self):
        m = ConcreteModel()
        m.x = Var(initialize=3.0)
        m.y = Var(initialize=2.0)
        exprA = m.x - m.y
        exprB = m.y - m.x

        # Nondeterministic form should not care which expression comes first
        # The nondeterministic form has been deprecated.
        #rep_A = generate_canonical_repn(exprA)
        #rep_B = generate_canonical_repn(exprB)
        #self.assertEqual( rep_A[-1], rep_B[-1] )
        
        # Deterministic form should care which expression comes first
        idMap_1 = {}
        rep_A = generate_canonical_repn(exprA, idMap_1)
        rep_B = generate_canonical_repn(exprB, idMap_1)
        self.assertEqual(set(map(lambda x:x.name, rep_A.variables)), set(map(lambda x:x.name, rep_B.variables)))
        
        idMap_2 = {}
        rep_B = generate_canonical_repn(exprB, idMap_2)
        rep_A = generate_canonical_repn(exprA, idMap_2)
        self.assertEqual(set(map(lambda x:x.name, rep_A.variables)), set(map(lambda x:x.name, rep_B.variables)))

        self.assertNotEqual( idMap_1, idMap_2 )
        
    def test_Expression_nonindexed(self):
        m = ConcreteModel()
        m.x = Var(initialize=3.0)
        m.y = Var(initialize=2.0)
        m.p = Param(initialize=1.0,mutable=True)
        m.e = Expression(initialize=1.0)

        # polynomial degree 0
        m.e.value = 0.0
        rep = generate_canonical_repn(m.e, {})
        rep1 = generate_canonical_repn(as_numeric(0.0), {})
        self.assertEqual(linear_repn_to_dict(rep1), linear_repn_to_dict(rep))

        rep = generate_canonical_repn(m.e+m.p, {})
        rep1 = generate_canonical_repn(0.0+m.p, {})
        self.assertEqual(linear_repn_to_dict(rep1), linear_repn_to_dict(rep))

        m.e.value = 0.0+m.p
        rep = generate_canonical_repn(m.e+m.p**2, {})
        rep1 = generate_canonical_repn(0.0+m.p+m.p**2, {})
        self.assertEqual(linear_repn_to_dict(rep1), linear_repn_to_dict(rep))

        # polynomial degree 1
        m.e.value = m.x
        rep = generate_canonical_repn(m.e, {})
        rep1 = generate_canonical_repn(m.x, {})
        self.assertEqual(linear_repn_to_dict(rep1), linear_repn_to_dict(rep))

        rep = generate_canonical_repn(m.e+m.y, {})
        rep1 = generate_canonical_repn(m.x+m.y, {})
        self.assertEqual(linear_repn_to_dict(rep1), linear_repn_to_dict(rep))

        m.e.value = 0.0
        rep = generate_canonical_repn(m.x+m.e, {})
        rep1 = generate_canonical_repn(m.x+0.0, {})
        self.assertEqual(linear_repn_to_dict(rep1), linear_repn_to_dict(rep))

        m.e.value = 1.0
        rep = generate_canonical_repn(m.x*m.e, {})
        rep1 = generate_canonical_repn(m.x*1.0, {})
        self.assertEqual(linear_repn_to_dict(rep1), linear_repn_to_dict(rep))

        # polynomial degree > 1
        m.e.value = m.x**2
        rep = generate_canonical_repn(m.e, {})
        rep1 = generate_canonical_repn(m.x**2, {})
        self.assertEqual(rep1, rep)

        rep = generate_canonical_repn(m.e+cos(m.y), {})
        rep1 = generate_canonical_repn(m.x**2+cos(m.y), {})
        self.assertEqual(rep1, rep)

        rep = generate_canonical_repn(m.e+cos(m.e), {})
        rep1 = generate_canonical_repn(m.x**2+cos(m.x**2), {})
        self.assertEqual(rep1, rep)

    def test_Expression_indexed(self):
        m = ConcreteModel()
        m.x = Var(initialize=3.0)
        m.y = Var(initialize=2.0)
        m.p = Param(initialize=1.0,mutable=True)
        m.e = Expression([1],initialize=1.0)

        # polynomial degree 0
        m.e[1].value = 0.0
        rep = generate_canonical_repn(m.e[1], {})
        rep1 = generate_canonical_repn(as_numeric(0.0), {})
        self.assertEqual(linear_repn_to_dict(rep1), linear_repn_to_dict(rep))

        rep = generate_canonical_repn(m.e[1]+m.p, {})
        rep1 = generate_canonical_repn(0.0+m.p, {})
        self.assertEqual(linear_repn_to_dict(rep1), linear_repn_to_dict(rep))

        m.e[1].value = 0.0+m.p
        rep = generate_canonical_repn(m.e[1]+m.p**2, {})
        rep1 = generate_canonical_repn(0.0+m.p+m.p**2, {})
        self.assertEqual(linear_repn_to_dict(rep1), linear_repn_to_dict(rep))

        # polynomial degree 1
        m.e[1].value = m.x
        rep = generate_canonical_repn(m.e[1], {})
        rep1 = generate_canonical_repn(m.x, {})
        self.assertEqual(linear_repn_to_dict(rep1), linear_repn_to_dict(rep))

        rep = generate_canonical_repn(m.e[1]+m.y, {})
        rep1 = generate_canonical_repn(m.x+m.y, {})
        self.assertEqual(linear_repn_to_dict(rep1), linear_repn_to_dict(rep))

        m.e[1].value = 0.0
        rep = generate_canonical_repn(m.x+m.e[1], {})
        rep1 = generate_canonical_repn(m.x+0.0, {})
        self.assertEqual(linear_repn_to_dict(rep1), linear_repn_to_dict(rep))

        m.e[1].value = 1.0
        rep = generate_canonical_repn(m.x*m.e[1], {})
        rep1 = generate_canonical_repn(m.x*1.0, {})
        self.assertEqual(linear_repn_to_dict(rep1), linear_repn_to_dict(rep))

        # polynomial degree > 1
        m.e[1].value = m.x**2
        rep = generate_canonical_repn(m.e[1], {})
        rep1 = generate_canonical_repn(m.x**2, {})
        self.assertEqual(rep1, rep)

        rep = generate_canonical_repn(m.e[1]+cos(m.y), {})
        rep1 = generate_canonical_repn(m.x**2+cos(m.y), {})
        self.assertEqual(rep1, rep)

        rep = generate_canonical_repn(m.e[1]+cos(m.e[1]), {})
        rep1 = generate_canonical_repn(m.x**2+cos(m.x**2), {})
        self.assertEqual(rep1, rep)

    def test_Expr_if_constant(self):
        model = ConcreteModel()
        model.x = Var()
        model.x.fix(2.0)

        rep = generate_canonical_repn(Expr_if(IF=model.x, THEN=1, ELSE=-1))
        self.assertTrue(isinstance(rep, LinearCanonicalRepn) == True)
        self.assertTrue(rep.linear == None)
        self.assertTrue(rep.constant != None)
        self.assertTrue(rep.variables == None)
        baseline = {  None        : 1}
        self.assertEqual(baseline,
                         linear_repn_to_dict(rep))
        rep = generate_canonical_repn(Expr_if(IF=model.x**2, THEN=1, ELSE=-1))
        self.assertTrue(isinstance(rep, LinearCanonicalRepn) == True)
        self.assertTrue(rep.linear == None)
        self.assertTrue(rep.constant != None)
        self.assertTrue(rep.variables == None)
        baseline = {  None        : 1}
        self.assertEqual(baseline,
                         linear_repn_to_dict(rep))
        rep = generate_canonical_repn(Expr_if(IF=(1-cos(model.x-1)) > 0.5, THEN=1, ELSE=-1))
        self.assertTrue(isinstance(rep, LinearCanonicalRepn) == True)
        self.assertTrue(rep.linear == None)
        self.assertTrue(rep.constant != None)
        self.assertTrue(rep.variables == None)
        baseline = {  None        : -1}
        self.assertEqual(baseline,
                         linear_repn_to_dict(rep))
        rep = generate_canonical_repn(Expr_if(IF=1, THEN=model.x, ELSE=-1))
        self.assertTrue(isinstance(rep, LinearCanonicalRepn) == True)
        self.assertTrue(rep.linear == None)
        self.assertTrue(rep.constant != None)
        self.assertTrue(rep.variables == None)
        baseline = {  None        : value(model.x)}
        self.assertEqual(baseline,
                         linear_repn_to_dict(rep))
        rep = generate_canonical_repn(Expr_if(IF=0, THEN=1, ELSE=model.x))
        self.assertTrue(isinstance(rep, LinearCanonicalRepn) == True)
        self.assertTrue(rep.linear == None)
        self.assertTrue(rep.constant != None)
        self.assertTrue(rep.variables == None)
        baseline = {  None        : value(model.x)}
        self.assertEqual(baseline,
                         linear_repn_to_dict(rep))

    def test_Expr_if_linear(self):
        model = ConcreteModel()
        model.x = Var()
        model.y = Var()

        rep = generate_canonical_repn(Expr_if(IF=1, THEN=model.x+3*model.y+10, ELSE=-1))
        self.assertTrue(isinstance(rep, LinearCanonicalRepn) == True)
        self.assertTrue(rep.linear != None)
        self.assertTrue(rep.constant != None)
        self.assertTrue(rep.variables != None)
        baseline = { id(model.x) : 1, 
                     id(model.y) : 3,
                     None        : 10}
        self.assertEqual(baseline,
                         linear_repn_to_dict(rep))

        rep = generate_canonical_repn(Expr_if(IF=0.0, THEN=1.0, ELSE=-model.x))
        self.assertTrue(isinstance(rep, LinearCanonicalRepn) == True)
        self.assertTrue(rep.linear != None)
        self.assertTrue(rep.constant == None)
        self.assertTrue(rep.variables != None)
        baseline = { id(model.x) : -1}
        self.assertEqual(baseline,
                         linear_repn_to_dict(rep))


    def test_Expr_if_quadratic(self):
        model = ConcreteModel()
        model.x = Var()

        rep = generate_canonical_repn(Expr_if(IF=1.0, THEN=model.x**2, ELSE=-1.0))
        self.assertTrue(isinstance(rep, GeneralCanonicalRepn) == True)
        self.assertEqual(canonical_degree(rep), 2)
        rep = generate_canonical_repn(Expr_if(IF=0.0, THEN=1.0, ELSE=-model.x**2))
        self.assertTrue(isinstance(rep, GeneralCanonicalRepn) == True)
        self.assertEqual(canonical_degree(rep), 2)

    def test_Expr_if_nonlinear(self):
        model = ConcreteModel()
        model.x = Var()

        rep = generate_canonical_repn(Expr_if(IF=model.x, THEN=1.0, ELSE=-1.0))
        self.assertTrue(isinstance(rep, GeneralCanonicalRepn) == True)
        self.assertEqual(canonical_degree(rep), None)
        rep = generate_canonical_repn(Expr_if(IF=1.0,
                                              THEN=Expr_if(IF=model.x**2, THEN=1.0, ELSE=-1.0),
                                              ELSE=-1.0))
        self.assertTrue(isinstance(rep, GeneralCanonicalRepn) == True)
        self.assertEqual(canonical_degree(rep), None)
        rep = generate_canonical_repn(Expr_if(IF=model.x**2, THEN=1.0, ELSE=-1.0))
        self.assertTrue(isinstance(rep, GeneralCanonicalRepn) == True)
        self.assertEqual(canonical_degree(rep), None)

    def test_Expr_if_nonlinear(self):
        model = ConcreteModel()
        model.x = Var()

        rep = generate_canonical_repn(Expr_if(IF=model.x, THEN=1.0, ELSE=-1.0))
        self.assertTrue(isinstance(rep, GeneralCanonicalRepn) == True)
        self.assertEqual(canonical_degree(rep), None)
        rep = generate_canonical_repn(Expr_if(IF=1.0,
                                              THEN=Expr_if(IF=model.x**2, THEN=1.0, ELSE=-1.0),
                                              ELSE=-1.0))
        self.assertTrue(isinstance(rep, GeneralCanonicalRepn) == True)
        self.assertEqual(canonical_degree(rep), None)
        rep = generate_canonical_repn(Expr_if(IF=model.x**2, THEN=1.0, ELSE=-1.0))
        self.assertTrue(isinstance(rep, GeneralCanonicalRepn) == True)
        self.assertEqual(canonical_degree(rep), None)

if __name__ == "__main__":
    unittest.main()




from coopr.pyomo import *

# Creates a list of the Ingredients
Ingredients = ['CHICKEN', 'BEEF', 'MUTTON', 'RICE', 'WHEAT', 'GEL']

# A dictionary of the costs of each of the Ingredients is created
costs = {'CHICKEN': 0.013,
         'BEEF': 0.008,
         'MUTTON': 0.010,
         'RICE': 0.002,
         'WHEAT': 0.005,
         'GEL': 0.001}

# A dictionary of the protein percent in each of the Ingredients is created
proteinPercent = {'CHICKEN': 0.100,
                  'BEEF': 0.200,
                  'MUTTON': 0.150,
                  'RICE': 0.000,
                  'WHEAT': 0.040,
                  'GEL': 0.000}

# A dictionary of the fat percent in each of the Ingredients is created
fatPercent = {'CHICKEN': 0.080,
              'BEEF': 0.100,
              'MUTTON': 0.110,
              'RICE': 0.010,
              'WHEAT': 0.010,
              'GEL': 0.000}

# A dictionary of the fibre percent in each of the Ingredients is created
fibrePercent = {'CHICKEN': 0.001,
                'BEEF': 0.005,
                'MUTTON': 0.003,
                'RICE': 0.100,
                'WHEAT': 0.150,
                'GEL': 0.000}

# A dictionary of the salt percent in each of the Ingredients is created
saltPercent = {'CHICKEN': 0.002,
               'BEEF': 0.005,
               'MUTTON': 0.007,
               'RICE': 0.002,
               'WHEAT': 0.008,
               'GEL': 0.000}

model = ConcreteModel(name="The Whiskas Problem")

model.ingredient_vars = Var(Ingredients, bounds=(0,None), doc="The amount of each ingredient that is used")

model.obj = Objective(expr=sum(costs[i]*model.ingredient_vars[i] for i in Ingredients), doc="Total Cost of Ingredients per can")

model.c0 = Constraint(expr=sum(model.ingredient_vars[i] for i in Ingredients) == 100, doc="PercentagesSum")
model.c1 = Constraint(expr=sum(proteinPercent[i] * model.ingredient_vars[i] for i in Ingredients) >= 8.0, doc="ProteinRequirement")
model.c2 = Constraint(expr=sum(fatPercent[i] * model.ingredient_vars[i] for i in Ingredients) >= 6.0, doc="FatRequirement")
model.c3 = Constraint(expr=sum(fibrePercent[i] * model.ingredient_vars[i] for i in Ingredients) <= 2.0, doc="FibreRequirement")
model.c4 = Constraint(expr=sum(saltPercent[i] * model.ingredient_vars[i] for i in Ingredients) <= 0.4, doc="SaltRequirement")



from pyutilib.misc import Options
from coopr.pyomo import *
import math
import random

def print_model_stats(options,model):
    print "-"*40
    if options is None:
        print "DEFAULT"
    else:
        print options.type
    rowc = {}
    for i in model.I:
        rowc[i] = 0
    colc = {}
    for i in model.J:
        colc[i] = 0
    for (i,j) in model.S:
            rowc[i] += 1        
            colc[j] += 1        
    print "Row Counts"
    s = 0.0
    for i in sorted(rowc):
        #print i, rowc[i]
        s += rowc[i]
    print "Average:",s/len(rowc)
    print "Col Counts"
    s = 0.0
    for i in sorted(colc):
        #print i, colc[i]
        s += colc[i]
    print "Average:",s/len(colc)
    print "I",len(model.I)
    print "J",len(model.J)
    print "-"*40

def pyomo_create_model(options=None, model_options=None):
    if model_options is None:
        model_options = Options()
    if model_options.type is None:
        model_options.type = 'fixed_set_size'
    #
    # m - number of elements
    #
    m = 100 if model_options.m is None else model_options.m
    #
    # n - number of sets
    #
    n = 200 if model_options.n is None else model_options.n
    seed = 9090 if model_options.seed is None else model_options.seed
    random.seed(9090)
    #
    if model_options.type == 'fixed_set_size':
        #
        # p   - fixed number elements per set
        # rho - fixed fraction of elements per set
        #
        p = model_options.p
        if p is None:
            if model_options.rho is None:
                p = int(math.ceil(m * 0.7))
            else:
                p = int(math.ceil(m * model_options.rho))
        #
        def S_rule(model):
            ans = set()
            for j in xrange(1,n+1):
                tmp = range(1,m+1)
                random.shuffle( tmp )
                for i in range(0,p):
                    ans.add( (tmp[i], j) )
            return ans
    elif model_options.type == 'fixed_element_coverage':
        #
        # p   - fixed number of sets that cover each element
        # rho - fixed fraction of sets that cover each element
        #
        p = model_options.p
        if p is None:
            if model_options.rho is None:
                p = int(math.ceil(n * 0.4))
            else:
                p = int(math.ceil(n * model_options.rho))
        #
        def S_rule(model):
            ans = set()
            for i in xrange(1,m+1):
                tmp = range(1,n+1)
                random.shuffle( tmp )
                for j in range(0,p):
                    ans.add( (i, tmp[j]) )
            return ans
    elif model_options.type == 'fixed_probability':
        #
        # rho - probability of selecting element for a set
        #
        rho = 0.3 if model_options.rho is None else model_options.rho
        #
        def S_rule(model):
            ans = set()
            for j in xrange(1,n+1):
                for i in xrange(1,m+1):
                    if random.uniform(0,1) < rho:
                        ans.add( (i, j) )
            return ans
    elif model_options.type == 'fixed_fill':
        #
        # rho - |S|/(I*J)
        #
        rho = 0.3 if model_options.rho is None else model_options.rho
        #
        def S_rule(model):
            ans = set()
            for j in xrange(1,n+1):
                for i in xrange(1,m+1):
                    if random.uniform(0,1) < rho:
                        ans.add( (i, j) )
            return ans
    #
    # CREATE MODEL
    #
    model = ConcreteModel()
    #
    # (i,j) in S if element i in set j
    #
    model.S = Set(dimen=2, initialize=S_rule)
    #
    # Dynamically create the I and J index sets, since
    # some rows or columns of S may not be populated.
    #
    def I_rule(model):
        return set((i for (i,j) in model.S))
    model.I = Set(initialize=I_rule)
    def J_rule(model):
        return set((j for (i,j) in model.S))
    model.J = Set(initialize=J_rule)
    #
    # Weights
    #
    model.w = Param(model.J, within=NonNegativeReals, initialize=1.0)
    #
    # Set selection binary variables
    #
    model.x = Var(model.J, within=Binary)
    #
    # Objective
    #
    def cost_rule(model):
        return summation(model.w, model.x)

    model.cost = Objective()

    #
    # Constraint
    #
    def cover_rule(model, i):
        expr = 0
        for j in model.x:
            if (i,j) in model.S:
                expr += model.x[j]
        #
        # WEH - this check is not needed, since I is constructed dynamically
        #
        #if expr is 0:
            #return Constraint.Skip
        return expr >= 1

    model.cover = Constraint(model.I)

    #
    print_model_stats(model_options, model)
    return model

def test_model(options=None):
    model = pyomo_create_model(model_options=options)
    #print_model_stats(options, model)

if __name__ == '__main__':
    test_model()
    #
    options = Options()
    options.type = 'fixed_set_size'
    options.m = 11
    options.n = 21
    options.rho = 0.3
    test_model(options)
    #
    options = Options()
    options.type = 'fixed_element_coverage'
    test_model(options)
    #
    options = Options()
    options.m = 100
    options.n = 200
    options.type = 'fixed_probability'
    test_model(options)
    #
    options = Options()
    options.type = 'fixed_element_coverage'
    options.m = 10
    options.n = 100
    options.rho = 0.1
    test_model(options)
    #




class ModelScript(object):

    def __init__(self, filename=None, text=None):
        if filename is not None:
            self._filename = filename
        elif text is not None:
            self._filename = None
            self._text = text
        else:
            raise ValueError("Must provide either a script file or text data")

    def read(self):
        if self._filename is not None:
            with open(self._filename, 'r') as f:
                return f.read()
        else:
            return self._text

    def filename(self):
        if self._filename is not None:
            return self._filename
        else:
            return "<unknown>"
